import functools
import operator

import pytest
from tqdm import tqdm

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=4
        ),
    ],
)
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 13],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 20, 32, 64],
)
@pytest.mark.parametrize(
    "pop_size",
    [0, 1, 2, 8, 13],
)
def test_pop_to_assemblage(
    retention_policy,
    num_deposits,
    differentia_bit_width,
    pop_size,
):
    pop = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_bit_width,
        ).CloneNthDescendant(num_deposits)
        for __ in range(pop_size)
    ]

    hsa = hstrat.pop_to_assemblage(
        pop, progress_wrap=functools.partial(tqdm, disable=True)
    )
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(
                col.IterRetainedRanks(),
                assemblage_specimen.GetData().notna().index,
            ),
        )
        for col, assemblage_specimen in zip(pop, hsa.BuildSpecimens())
    )
    assert all(
        map(
            lambda x: operator.eq(*x),
            zip(
                col.IterRetainedDifferentia(),
                assemblage_specimen.GetData().notna(),
            ),
        )
        for col, assemblage_specimen in zip(pop, hsa.BuildSpecimens())
    )



import typing

import numpy as np

from ._jit import jit
from ._reversed_enumerate import reversed_enumerate

_reversed_enumerate_jit = jit(nopython=True)(reversed_enumerate)

# adapted from https://stackoverflow.com/a/41578614
@jit(nopython=True)
def numpy_index_flat(
    array: np.ndarray,
    item: typing.Any,
    n: int = 0,
) -> typing.Optional[int]:
    """Find the index of the nth (zero-indexed) occurrence of a given item in a
    Numpy array.

    Parameters
    ----------
    array : numpy.ndarray
        The Numpy array to search.
    item : Any
        The item to search for in the array.
    n : int, default 0
        Specifies the zero-indexed nth occurrence of `item` to be
        found.

        If not specified or set to 0, the function returns the index of
        the first occurrence of `item`. If there are fewer than `n` occurrences
        of `item` in the array, the function returns None.

        If `n` is negative,
        the function returns the index of the nth occurrence from the end of
        the array (i.e., -1 returns the last occurrence, -2 returns the
        second-to-last occurrence, and so on).

    Returns
    -------
    int or None
        The index of the nth (zero-indexed) occurrence of the given item in the
        array, or None if the item is not found in the array.

    Examples
    --------
    >>> a = np.array([1, 4, 3, 4, 5])
    >>> numpy_index_flat(a, 3)
    2,
    >>> numpy_index_flat(a, 6)
    None
    >>> numpy_index_flat(a, 3, n=-1)
    3
    """
    # numba can't handle coexisting type possibilities for enumeration;
    # must spell out separately
    if n < 0:
        enumeration = _reversed_enumerate_jit(array)
        for idx, val in enumeration:
            if val == item:
                n += 1
            if n == 0:
                return idx
        return None

    else:
        enumeration = enumerate(array)
        for idx, val in enumeration:
            if val == item:
                n -= 1
            if n == -1:
                return idx
        return None



import pandas as pd

from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids


def alifestd_count_leaf_nodes(phylogeny_df: pd.DataFrame) -> int:
    """How many leaf nodes are contained in phylogeny?"""
    return len(alifestd_find_leaf_ids(phylogeny_df))



import pandas as pd


def alifestd_convert_root_ancestor_token(
    ancestor_list: pd.Series,
    root_ancestor_token: str,
    mutate: bool = False,
) -> pd.Series:
    """Set `root_ancestor_token` for ancestor_list series.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    """

    if not mutate:
        ancestor_list = ancestor_list.copy()

    ancestor_list[
        ancestor_list.str.lower().isin(("[none]", "[]"))
    ] = f"[{root_ancestor_token}]"
    return ancestor_list



import typing

from ...._detail import PolicyCouplerBase
from ..._PolicySpec import PolicySpec
from ..._impl import calc_provided_uncertainty


class IterRetainedRanksProduction:
    """Functor to provide member function implementation in Policy class.

    Optimizes to reduce calls to `calc_provided_uncertainty`.
    """

    def __init__(
        self: "IterRetainedRanksProduction",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "IterRetainedRanksProduction", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "IterRetainedRanksProduction",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> typing.Iterator[int]:
        """Iterate over retained strata ranks at `num_strata_deposited` in
        ascending order."""
        spec = policy.GetSpec()
        resolution = spec.GetRecencyProportionalResolution()

        cur_rank = 0
        last_rank = num_strata_deposited - 1
        while last_rank - cur_rank > resolution:
            yield cur_rank
            cur_rank += calc_provided_uncertainty(
                resolution,
                last_rank - cur_rank,
            )

        yield from range(
            max(last_rank - resolution, 0),
            num_strata_deposited,
        )



import typing

from iterpop import iterpop as ip

from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_to_records import col_to_records


def pop_to_records(
    columns: typing.Iterable[HereditaryStratigraphicColumn],
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.Dict:
    """Serialize a sequence of `HereditaryStratigraphicColumn`s to a dict
    composed of builtin types.

    Parameters
    ----------
    columns : iterable of HereditaryStratigraphicColumn
        Data to serialize.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    """
    col_records = [col_to_records(column) for column in progress_wrap(columns)]

    res = {}
    for common_field in (
        "policy",
        "policy_algo",
        "policy_spec",
        "differentia_bit_width",
        "hstrat_version",
    ):
        res[common_field] = ip.pourhomogeneous(
            col_record.pop(common_field) for col_record in col_records
        )

    res["columns"] = col_records

    return res



import functools
import typing

import pandas as pd
import polars as pl

from hstrat._auxiliary_lib import coerce_to_pandas, coerce_to_polars


def enforce_identical_polars_result(
    func: typing.Callable, recurse_type_checks: bool = False
) -> typing.Callable:
    """
    Decorates `func`, which must accept both polars- and pandas-types arguments
    (i.e., wrapped with `_auxiliary_lib.delegate_polars_implementation`) to
    test that results are identical whether polars or pandas arguments are used.
    """

    @functools.wraps(func)
    def polars_equality_checker(*args, **kwargs) -> typing.Any:
        pl_args = tuple(
            coerce_to_polars(arg, recurse=recurse_type_checks) for arg in args
        )
        pl_kwargs = {
            kw: coerce_to_polars(arg, recurse=recurse_type_checks)
            for kw, arg in kwargs.items()
        }

        pd_result = func(*args, **kwargs)
        pl_result = func(*pl_args, **pl_kwargs)
        if isinstance(pl_result, pl.DataFrame):
            pl_result = coerce_to_pandas(
                pl_result, recurse=recurse_type_checks
            )
            pd.testing.assert_frame_equal(pl_result, pd_result)
        elif isinstance(pl_result, pl.Series):
            pl_result = coerce_to_pandas(
                pl_result, recurse=recurse_type_checks
            )
            pd.testing.assert_series_equal(
                pl_result, pd_result, check_names=False
            )
        else:
            assert pd_result == pl_result
        return pd_result

    return polars_equality_checker



import typing
import warnings

import alifedata_phyloinformatics_convert as apc
import opytional as opyt
import pandas as pd

from ...._auxiliary_lib import (
    BioPhyloTree,
    alifestd_find_root_ids,
    alifestd_make_empty,
    alifestd_validate,
)
from ....genome_instrumentation import HereditaryStratigraphicColumn
from ...population import build_distance_matrix_biopython
from ._append_genesis_organism import append_genesis_organism
from ._time_calibrate_tree import time_calibrate_tree


def build_tree_biopython_distance(
    algo: str,
    population: typing.Sequence[HereditaryStratigraphicColumn],
    estimator: str,
    prior: typing.Union[str, typing.Any],
    taxon_labels: typing.Optional[typing.Iterable],
    force_common_ancestry: bool,
    negative_origin_time_correction_method: typing.Optional[str],
) -> pd.DataFrame:
    """Backend interface to biopython distance-based tree reconstruction
    methods."""

    # biopython doesn't represent empty tree elegantly
    # for simplicity, return early for this special case
    if len(population) == 0:
        return alifestd_make_empty()

    taxon_labels = list(
        opyt.or_value(
            taxon_labels,
            map(str, range(len(population))),
        )
    )

    distance_matrix = build_distance_matrix_biopython(
        population,
        estimator,
        prior,
        taxon_labels,
        force_common_ancestry or None,
    )
    constructor = getattr(BioPhyloTree.DistanceTreeConstructor(), algo)
    biopython_tree = (
        BioPhyloTree.BaseTree.Tree()
        if not distance_matrix
        else constructor(distance_matrix)
    )

    # convert, calibrate, and return
    alifestd_df = apc.biopython_tree_to_alife_dataframe(biopython_tree)
    alifestd_df["taxon_label"] = alifestd_df["name"]

    min_edge_length = alifestd_df["branch_length"].min()
    if min_edge_length < -1e-12:  # why? pytest approx absolute tolerance
        warnings.warn(
            f"""Negative branch length(s) estimated with minimum {
                min_edge_length
            }; returning estimated tree without further rerooting """
            "or origin time estimation. "
        )
        return alifestd_df

    # set near-zero negative branch lengths to zero
    alifestd_df["branch_length"] = alifestd_df["branch_length"].clip(lower=0)

    id_lookup = dict(zip(alifestd_df["taxon_label"], alifestd_df["id"]))
    col_lookup = dict(zip(taxon_labels, population))

    leaf_origin_times = {
        id_lookup[taxon_label]: col_lookup[taxon_label].GetNumStrataDeposited()
        - 1
        for taxon_label in taxon_labels
    }
    alifestd_df = time_calibrate_tree(
        alifestd_df, leaf_origin_times, negative_origin_time_correction_method
    )

    alifestd_df = append_genesis_organism(alifestd_df, mutate=True)

    assert alifestd_validate(alifestd_df)
    assert len(alifestd_find_root_ids(alifestd_df))

    return alifestd_df



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import calc_provided_uncertainty


class CalcMrcaUncertaintyAbsExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcMrcaUncertaintyAbsExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> int:
        """Exactly how much uncertainty to estimate rank of MRCA?"""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        spec = policy.GetSpec()
        guaranteed_resolution = spec.GetDepthProportionalResolution()

        least_num_strata_deposited = min(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )
        least_last_rank = least_num_strata_deposited - 1

        # mrca at very last rank
        if actual_rank_of_mrca == least_last_rank:
            return 0
        # haven't added enough ranks to start dropping ranks
        elif least_last_rank < guaranteed_resolution * 2:
            return 0

        cur_stage_uncertainty = calc_provided_uncertainty(
            guaranteed_resolution,
            least_num_strata_deposited,
        )
        cur_stage_max_idx = (  # noqa: F841, keep unused for comprehensibility
            least_num_strata_deposited // cur_stage_uncertainty
        )

        prev_stage_uncertainty = cur_stage_uncertainty // 2
        prev_stage_max_idx = (least_last_rank - 1) // prev_stage_uncertainty

        thresh_idx = (
            2 * prev_stage_max_idx - 4 * guaranteed_resolution + 2
        ) // 2

        # note that cur stage uncertainty is iterated through first
        # because ranks are removed from the back, the old prev stage
        # uncertainty lingers at more recent ranks
        if actual_rank_of_mrca < thresh_idx * cur_stage_uncertainty:
            # mrca between last regularly-spaced rank and tail rank
            if actual_rank_of_mrca >= (
                least_last_rank - least_last_rank % cur_stage_uncertainty
            ):
                return (least_last_rank - 1) % cur_stage_uncertainty
            # mrca between two regularly-spaced ranks
            else:
                return cur_stage_uncertainty - 1
        # mrca between last regularly-spaced rank and tail rank
        elif actual_rank_of_mrca >= (
            least_last_rank - least_last_rank % prev_stage_uncertainty
        ):
            return (least_last_rank - 1) % prev_stage_uncertainty
        # mrca between two regularly-spaced ranks
        else:
            return prev_stage_uncertainty - 1



from ._PolicyCouplerBase import PolicyCouplerBase
from ._PolicyCouplerFactory import PolicyCouplerFactory
from ._PolicySpecBase import PolicySpecBase
from ._UnsatisfiableParameterizationRequestError import (
    UnsatisfiableParameterizationRequestError,
)

__all__ = [
    "PolicyCouplerBase",
    "PolicyCouplerFactory",
    "PolicySpecBase",
    "UnsatisfiableParameterizationRequestError",
]



from ._FromPredKeepRank import FromPredKeepRank
from ._GenDropRanks import GenDropRanks
from ._GenDropRanksNaive import GenDropRanksNaive

_GenDropRanks_impls = (FromPredKeepRank, GenDropRanks, GenDropRanksNaive)



import os

import alifedata_phyloinformatics_convert as apc
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_find_leaf_ids,
    alifestd_is_asexual,
    alifestd_parse_ancestor_ids,
    alifestd_to_working_format,
    alifestd_try_add_ancestor_id_col,
    swap_rows_and_indices,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_find_leaf_ids_empty(phylogeny_df, apply):
    phylogeny_df = phylogeny_df.copy()
    phylogeny_df = apply(phylogeny_df)
    assert alifestd_find_leaf_ids(phylogeny_df.iloc[-1:0, :]) == []


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_find_leaf_ids_singleton(phylogeny_df, apply):
    phylogeny_df = phylogeny_df.copy()
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df.sort_values("id", ascending=True, inplace=True)

    assert alifestd_find_leaf_ids(phylogeny_df.iloc[0:1, :]) == [
        phylogeny_df.iloc[0].at["id"]
    ]


def test_alifestd_find_leaf_ids_tworoots():
    phylo1 = pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    phylo1.sort_values("id", ascending=True, inplace=True)

    phylo2 = pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
    phylo2.sort_values("id", ascending=True, inplace=True)
    phylo2["id"] += phylo1["id"].max()

    assert alifestd_find_leaf_ids(
        pd.concat(
            [
                phylo1.iloc[0:1, :],
                phylo2.iloc[0:1, :],
            ]
        )
    ) == [phylo1.iloc[0].at["id"]] + [phylo2.iloc[0].at["id"]]


def test_alifestd_find_leaf_ids_empty2():
    phylo1 = pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    assert alifestd_find_leaf_ids(phylo1[-1:0]) == []
    phylo1["ancestor_id"] = 0
    assert alifestd_find_leaf_ids(phylo1[-1:0]) == []


def _test_alifestd_find_leaf_ids_impl(phylogeny_df):
    phylogeny_df_ = phylogeny_df.set_index("id")
    if alifestd_is_asexual(phylogeny_df):
        trees = apc.alife_dataframe_to_dendropy_trees(phylogeny_df)
        leaf_ids = [
            leaf_node.id
            for tree in trees
            for leaf_node in tree.leaf_node_iter()
        ]
        leaf_ids.sort(key=phylogeny_df_.index.get_loc)

        assert leaf_ids == alifestd_find_leaf_ids(phylogeny_df)
    else:
        # sexual phylogenies
        leaf_ids = alifestd_find_leaf_ids(phylogeny_df)
        assert sorted(leaf_ids, key=phylogeny_df_.index.get_loc) == leaf_ids

        all_ids = set(phylogeny_df["id"])
        internal_ids = all_ids - set(leaf_ids)

        for leaf_id in leaf_ids:
            assert not any(
                leaf_id in alifestd_parse_ancestor_ids(ancestor_list_str)
                for ancestor_list_str in phylogeny_df["ancestor_list"]
            )
        for internal_id in internal_ids:
            assert any(
                internal_id in alifestd_parse_ancestor_ids(ancestor_list_str)
                for ancestor_list_str in phylogeny_df["ancestor_list"]
            )


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_find_leaf_ids_twolineages(phylogeny_df, apply):
    phylogeny_df = phylogeny_df.copy()
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df.sort_values("id", ascending=True, inplace=True)
    phylogeny_df.reset_index(inplace=True)

    aggregated_df = alifestd_aggregate_phylogenies(
        [
            phylogeny_df.copy(),
            phylogeny_df.copy(),
        ]
    )

    lineage1 = aggregated_df[len(phylogeny_df) :]
    lineage2 = aggregated_df[: len(phylogeny_df)]

    _test_alifestd_find_leaf_ids_impl(
        pd.concat(
            [
                lineage1,
                lineage2,
            ]
        )
    )
    _test_alifestd_find_leaf_ids_impl(
        pd.concat(
            [
                lineage1,
                lineage2,
            ]
        ).sort_index()
    )

    _test_alifestd_find_leaf_ids_impl(
        pd.concat(
            [
                lineage1.iloc[::-1, :],
                lineage2,
            ]
        )
    )
    _test_alifestd_find_leaf_ids_impl(
        pd.concat(
            [
                lineage1.iloc[:10:-1, :],
                lineage2,
                lineage1.iloc[10::-1, :],
            ]
        )
    )
    _test_alifestd_find_leaf_ids_impl(
        pd.concat(
            [
                lineage1.iloc[10::, :],
                lineage2,
                lineage1.iloc[:10:, :],
            ]
        )
    )


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_find_leaf_ids_true(phylogeny_df, apply):
    phylogeny_df = phylogeny_df.copy()
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df.sort_values("id", ascending=True, inplace=True)

    phylogeny_df_ = phylogeny_df.copy()
    _test_alifestd_find_leaf_ids_impl(phylogeny_df)

    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_find_leaf_ids_false(phylogeny_df, apply):
    phylogeny_df = phylogeny_df.copy()
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df.sort_values("id", ascending=True, inplace=True)
    phylogeny_df_ = phylogeny_df.copy()
    _test_alifestd_find_leaf_ids_impl(phylogeny_df)
    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)

    phylogeny_df.sort_values("id", ascending=False, inplace=True)
    phylogeny_df.set_index("id", drop=False, inplace=True)

    phylogeny_df_ = phylogeny_df.copy()
    _test_alifestd_find_leaf_ids_impl(phylogeny_df)
    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)

    # one-by-one transpositions
    for idx, row in phylogeny_df.sample(min(10, len(phylogeny_df))).iterrows():
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            phylogeny_df_ = swap_rows_and_indices(
                phylogeny_df, idx, ancestor_id
            )
            _test_alifestd_find_leaf_ids_impl(phylogeny_df_)

    # cumulative transpositions in random order
    for idx, row in phylogeny_df.sample(min(10, len(phylogeny_df))).iterrows():
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            assert ancestor_id in phylogeny_df_.index
            phylogeny_df = swap_rows_and_indices(
                phylogeny_df, idx, ancestor_id
            )
            _test_alifestd_find_leaf_ids_impl(phylogeny_df)



import numpy as np

from ._detail import PriorBase


class UniformPrior(PriorBase):
    """Enacts a prior expectation that the generation of the most recent common
    ancestor (MRCA) is equally likely to occur at any generation since genesis.

    See Also
    --------
    GeometricPrior
        An exact, discrete analog of this prior.
    """

    def CalcIntervalProbabilityProxy(
        self: "UniformPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Characterize the prior probability of the MRCA generation falling
        within an interval range.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The proxy statistic, proportional to the true estimated interval
            probability of the MRCA value by a fixed (but unspecified) constant
            proportion.
        """
        return float(end_rank - begin_rank)

    def CalcIntervalConditionedMean(
        self: "UniformPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Calculate the centriod of prior probability mass within an interval
        of possible MRCA generations.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The prior expected generation of MRCA conditioned on the assumption
            that the MRCA falls within the given interval.
        """
        return np.mean((begin_rank, end_rank - 1))

    def SampleIntervalConditionedValue(
        self: "UniformPrior", begin_rank: int, end_rank: int
    ) -> int:
        """Sample a generation of the MRCA conditioned on the assumption that
        the MRCA falls within the given interval.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        int
            A sampled generation of the MRCA, conditioned on the assumption that
            the MRCA falls within the given interval.
        """
        raise NotImplementedError()



import pytest

from hstrat._auxiliary_lib import except_wrap_sentinel


def test_except_wrap_sentinel():
    def fail_division(a, b):
        assert a != 0
        return a / b

    # Wrap the function to return `float('inf')` on ZeroDivisionError
    safe_divide = except_wrap_sentinel(
        fail_division,
        {
            ZeroDivisionError: "Division by zero!",
            AssertionError: "Division with zero!",
        },
        sentinel=float("inf"),
    )

    # Test that dividing by zero returns sentinel and emits a warning
    with pytest.warns(UserWarning, match="Division by zero!"):
        assert safe_divide(1, 0) == float("inf")

    # Test that a different error type has its corresponding warning
    with pytest.warns(UserWarning, match="Division with zero!"):
        assert safe_divide(0, 2) == float("inf")

    # Test a successful division returns the correct result
    assert safe_divide(4, 2) == 2.0

    # Test that another exception not in errors is still raised
    with pytest.raises(TypeError):
        safe_divide("4", "2")



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_root_ids,
    alifestd_has_multiple_roots,
    alifestd_make_empty,
    alifestd_mark_root_id,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    original = phylogeny_df.copy()

    result = alifestd_mark_root_id(phylogeny_df)

    assert alifestd_validate(result)
    assert original.equals(phylogeny_df)

    assert not alifestd_has_multiple_roots(phylogeny_df)
    (root_id,) = alifestd_find_root_ids(phylogeny_df)
    assert all(result["root_id"] == root_id)


def test_empty():
    res = alifestd_mark_root_id(alifestd_make_empty())
    assert "root_id" in res
    assert len(res) == 0


def test_singleton():
    record = {"id": 0, "ancestor_id": 0, "ancestor_list": "[None]"}
    original = pd.DataFrame.from_records([record])

    res = alifestd_mark_root_id(original)
    assert alifestd_validate(res)
    assert res["root_id"].squeeze() == 0

    res = alifestd_mark_root_id(original.drop("ancestor_id", axis=1))
    assert alifestd_validate(res)
    assert res["root_id"].squeeze() == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_asexual1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [42, 1, 2],
            "ancestor_list": ["[None]", "[42]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_root_id(
        phylogeny_df,
        mutate=mutate,
    )
    assert (result_df["root_id"] == 42).all()

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_asexual2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 42, 2, 3],
            "ancestor_id": [42, 42, 42, 1],
            "ancestor_list": ["[42]", "[None]", "[42]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_root_id(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert (result_df["root_id"] == 42).all()

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_asexual3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_root_id(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "root_id"] == 1
    assert result_df.loc[0, "root_id"] == 0
    assert result_df.loc[2, "root_id"] == 0
    assert result_df.loc[3, "root_id"] == 1

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_sexual1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0, 1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_root_id(
        phylogeny_df,
        mutate=mutate,
    )
    assert (result_df["root_id"] == 0).all()

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_sexual2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 42, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[1, 42]", "[2, 42]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_root_id(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "root_id"] == 1
    assert result_df.loc[42, "root_id"] == 42
    assert result_df.loc[2, "root_id"] == 1
    assert result_df.loc[3, "root_id"] == 1

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_sexual3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3, 4],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]", "[1, 3]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_root_id(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "root_id"] == 1
    assert result_df.loc[0, "root_id"] == 0
    assert result_df.loc[2, "root_id"] == 0
    assert result_df.loc[3, "root_id"] == 1
    assert result_df.loc[4, "root_id"] == 1

    if not mutate:
        assert original_df.equals(phylogeny_df)



import pandas as pd
import pytest

from hstrat._auxiliary_lib import alifestd_convert_root_ancestor_token


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("token", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_empty(mutate, token):
    ancestor_list_series = pd.Series(dtype=str)
    expected_output = pd.Series(dtype=str)
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, token, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_singleton(
    mutate, tokenfrom, tokento
):
    ancestor_list_series = pd.Series([f"[{tokenfrom}]"])
    expected_output = pd.Series([f"[{tokento}]"])
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_single_root1(
    mutate, tokenfrom, tokento
):
    ancestor_list_series = pd.Series([f"[{tokenfrom}]", "[0]", "[0,1]"])
    expected_output = pd.Series([f"[{tokento}]", "[0]", "[0,1]"])
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_single_root2(
    mutate, tokenfrom, tokento
):
    ancestor_list_series = pd.Series(["[1]", f"[{tokenfrom}]", "[1]"])
    expected_output = pd.Series(["[1]", f"[{tokento}]", "[1]"])
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_multiple_roots1(
    mutate, tokenfrom, tokento
):
    ancestor_list_series = pd.Series(
        [f"[{tokenfrom}]", f"[{tokenfrom}]", "[1]"]
    )
    expected_output = pd.Series([f"[{tokento}]", f"[{tokento}]", "[1]"])
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_multiple_roots2(
    mutate, tokenfrom, tokento
):
    ancestor_list_series = pd.Series(
        [f"[{tokenfrom}]", f"[{tokenfrom}]", "[0, 1]"]
    )
    expected_output = pd.Series([f"[{tokento}]", f"[{tokento}]", "[0, 1]"])
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_multiple_roots3(
    mutate, tokenfrom, tokento
):
    ancestor_list_series = pd.Series(
        [f"[{tokenfrom}]", "[0, 1]", f"[{tokenfrom}]"]
    )
    expected_output = pd.Series([f"[{tokento}]", "[0, 1]", f"[{tokento}]"])
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_multiple_roots4(
    mutate, tokenfrom, tokento
):
    ancestor_list_series = pd.Series(
        [f"[{tokenfrom}]", f"[{tokenfrom}]", f"[{tokenfrom}]"]
    )
    expected_output = pd.Series(
        [f"[{tokento}]", f"[{tokento}]", f"[{tokento}]"]
    )
    output = alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=mutate
    )
    pd.testing.assert_series_equal(output, expected_output)


@pytest.mark.parametrize(
    "subject",
    [
        "pd.Series(dtype=str)",
        'pd.Series([f"[{tokenfrom}]"])',
        'pd.Series([f"[{tokenfrom}]", "[0]", "[0,1]"])',
        'pd.Series(["[1]", f"[{tokenfrom}]", "[1]"])',
        'pd.Series([f"[{tokenfrom}]", f"[{tokenfrom}]", "[1]"])',
        'pd.Series([f"[{tokenfrom}]", f"[{tokenfrom}]", "[0, 1]"])',
        'pd.Series([f"[{tokenfrom}]", f"[{tokenfrom}]", f"[{tokenfrom}]"])',
        'pd.Series([f"[{tokenfrom}]", "[0, 1]", f"[{tokenfrom}]"])',
    ],
)
@pytest.mark.parametrize("tokenfrom", ["none", "None", ""])
@pytest.mark.parametrize("tokento", ["none", "None", ""])
def test_alifestd_convert_root_ancestor_token_with_multiple_roots5(
    subject, tokenfrom, tokento
):
    ancestor_list_series = eval(subject)
    alifestd_convert_root_ancestor_token(
        ancestor_list_series, tokento, mutate=False
    )
    pd.testing.assert_series_equal(ancestor_list_series, eval(subject))



import anytree
import networkx as nx
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import (
    AnyTreeAscendingIter,
    anytree_cardinality,
    random_tree,
    seed_random,
)
import hstrat.phylogenetic_inference.tree._impl as impl
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.SampleAncestralRollbacksTriePostprocessor,
        detail.TriePostprocessorBase,
    )


# Test case for a trie with only one node
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize(
    "p_differentia_collision",
    [
        0,
        0.5,
        1.0,
    ],
)
def test_sample_ancestral_rollbacks_single_node_trie(
    mutate,
    p_differentia_collision,
):
    trie = impl.TrieInnerNode()
    root = hstrat.SampleAncestralRollbacksTriePostprocessor()(
        trie, p_differentia_collision=p_differentia_collision, mutate=mutate
    )
    assert root.parent is None
    # Since there is only one node in the trie, we don't expect any new nodes
    expected_num_nodes = 1
    assert anytree_cardinality(trie) == expected_num_nodes


@pytest.mark.parametrize("mutate", [True, False])
def test_sample_ancestral_rollbacks_multiple_levels_trie_phalf(mutate):
    root = impl.TrieInnerNode(rank=None, differentia=None)
    impl.TrieInnerNode(rank=0, differentia=101, parent=root)
    for i in range(5):
        node = impl.TrieInnerNode(
            rank=1, differentia=i, parent=root.children[0]
        )
        impl.TrieLeafNode(parent=node, taxon_label=f"{i}")
    another = impl.TrieLeafNode(parent=root, taxon_label="another")
    impl.TrieLeafNode(parent=another, taxon_label="another_leaf")

    # (rank None, diff None) @ KQ
    # ├── (rank 0, diff 101) @ MQ
    # │   ├── (rank 1, diff 0) @ NQ
    # │   │   └── 0 @ OQ
    # │   ├── (rank 1, diff 1) @ SQ
    # │   │   └── 1 @ IQ
    # │   ├── (rank 1, diff 2) @ BQ
    # │   │   └── 2 @ CQ
    # │   ├── (rank 1, diff 3) @ HQ
    # │   │   └── 3 @ FQ
    # │   └── (rank 1, diff 4) @ EQ
    # │       └── 4 @ nQ
    # └── another @ jQ
    #     └── another_leaf @ hQ

    original_num_nodes = anytree_cardinality(root)
    original_num_leaves = len(root.leaves)
    assert len(root.children) == 2

    before_ascending_content = [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]

    p_differentia_collision = 0.5
    root = hstrat.SampleAncestralRollbacksTriePostprocessor()(
        root, p_differentia_collision=p_differentia_collision, mutate=mutate
    )
    # for p 0.5, taking into account succession
    # doubles expected unzips from 2 to 4
    assert anytree_cardinality(root) >= original_num_nodes + 4
    assert len(root.leaves) == original_num_leaves
    assert len(root.children) == 6

    assert before_ascending_content == [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]


@pytest.mark.parametrize("mutate", [True, False])
def test_sample_ancestral_rollbacks_multiple_levels_trie_pzero(mutate):
    root = impl.TrieInnerNode(rank=None, differentia=None)
    impl.TrieInnerNode(rank=0, differentia=101, parent=root)
    for i in range(5):
        node = impl.TrieInnerNode(
            rank=1, differentia=i, parent=root.children[0]
        )
        impl.TrieLeafNode(parent=node, taxon_label=f"{i}")
    another = impl.TrieLeafNode(parent=root, taxon_label="another")
    impl.TrieLeafNode(parent=another, taxon_label="another_leaf")

    # (rank None, diff None) @ KQ
    # ├── (rank 0, diff 101) @ MQ
    # │   ├── (rank 1, diff 0) @ NQ
    # │   │   └── 0 @ OQ
    # │   ├── (rank 1, diff 1) @ SQ
    # │   │   └── 1 @ IQ
    # │   ├── (rank 1, diff 2) @ BQ
    # │   │   └── 2 @ CQ
    # │   ├── (rank 1, diff 3) @ HQ
    # │   │   └── 3 @ FQ
    # │   └── (rank 1, diff 4) @ EQ
    # │       └── 4 @ nQ
    # └── another @ jQ
    #     └── another_leaf @ hQ

    original_num_nodes = anytree_cardinality(root)
    original_num_leaves = len(root.leaves)
    assert len(root.children) == 2

    before_ascending_content = [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]

    p_differentia_collision = 0
    root = hstrat.SampleAncestralRollbacksTriePostprocessor()(
        root, p_differentia_collision=p_differentia_collision, mutate=mutate
    )
    assert anytree_cardinality(root) == original_num_nodes
    assert len(root.leaves) == original_num_leaves
    assert len(root.children) == 2

    assert before_ascending_content == [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]


@pytest.mark.parametrize("mutate", [True, False])
def test_sample_ancestral_rollbacks_multiple_levels_trie_pone(mutate):
    root = impl.TrieInnerNode(rank=None, differentia=None)
    impl.TrieInnerNode(rank=0, differentia=101, parent=root)
    for i in range(5):
        node = impl.TrieInnerNode(
            rank=1, differentia=i, parent=root.children[0]
        )
        impl.TrieLeafNode(parent=node, taxon_label=f"{i}")
    another = impl.TrieLeafNode(parent=root, taxon_label="another")
    impl.TrieLeafNode(parent=another, taxon_label="another_leaf")

    # (rank None, diff None) @ KQ
    # ├── (rank 0, diff 101) @ MQ
    # │   ├── (rank 1, diff 0) @ NQ
    # │   │   └── 0 @ OQ
    # │   ├── (rank 1, diff 1) @ SQ
    # │   │   └── 1 @ IQ
    # │   ├── (rank 1, diff 2) @ BQ
    # │   │   └── 2 @ CQ
    # │   ├── (rank 1, diff 3) @ HQ
    # │   │   └── 3 @ FQ
    # │   └── (rank 1, diff 4) @ EQ
    # │       └── 4 @ nQ
    # └── another @ jQ
    #     └── another_leaf @ hQ

    original_num_nodes = anytree_cardinality(root)
    original_num_leaves = len(root.leaves)
    assert len(root.children) == 2

    before_ascending_content = [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]

    p_differentia_collision = 1
    root = hstrat.SampleAncestralRollbacksTriePostprocessor()(
        root, p_differentia_collision=p_differentia_collision, mutate=mutate
    )
    assert anytree_cardinality(root) == original_num_nodes + 4
    assert len(root.leaves) == original_num_leaves
    assert len(root.children) == 6

    assert before_ascending_content == [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("tree_size", [10, 30, 100])
@pytest.mark.parametrize(
    "tree_seed",
    range(10),
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(2),
        hstrat.fixed_resolution_algo.Policy(2),
    ],
)
@pytest.mark.parametrize(
    "p_differentia_collision",
    [
        0,
        0.5,
        1.0,
        2**32,  # test full rollback
    ],
)
def test_unzip_fuzz(
    mutate, tree_seed, tree_size, retention_policy, p_differentia_collision
):

    seed_random(tree_seed)
    nx_tree = random_tree(n=tree_size, seed=tree_seed, create_using=nx.DiGraph)
    extant_population = hstrat.descend_template_phylogeny_networkx(
        nx_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ),
    )
    root = impl.build_trie_from_artifacts(
        extant_population,
        taxon_labels=None,
        progress_wrap=lambda x: x,
    )

    original_num_nodes = anytree_cardinality(root)
    original_num_leaves = len(root.leaves)

    before_ascending_content = [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]

    root = hstrat.SampleAncestralRollbacksTriePostprocessor()(
        root, p_differentia_collision=p_differentia_collision, mutate=mutate
    )
    assert (
        anytree_cardinality(root) > original_num_nodes
        or p_differentia_collision == 0
    )
    assert len(root.leaves) == original_num_leaves
    if p_differentia_collision > 1:
        assert len(root.children) == len(root.leaves)

    assert before_ascending_content == [
        [
            (node.taxon_label)
            if isinstance(node, impl.TrieLeafNode)
            else (node._rank, node._differentia)
            for node in AnyTreeAscendingIter(leaf)
        ]
        for leaf in sorted(root.leaves, key=lambda x: x.taxon_label)
    ]


def test_sample_ancestral_rollbacks_multiple_levels_seed():
    root = impl.TrieInnerNode(rank=None, differentia=None)
    impl.TrieInnerNode(rank=0, differentia=101, parent=root)
    for i in range(5):
        node = impl.TrieInnerNode(
            rank=1, differentia=i, parent=root.children[0]
        )
        impl.TrieLeafNode(parent=node, taxon_label=f"{i}")
    another = impl.TrieLeafNode(parent=root, taxon_label="another")
    impl.TrieLeafNode(parent=another, taxon_label="another_leaf")

    # (rank None, diff None) @ KQ
    # ├── (rank 0, diff 101) @ MQ
    # │   ├── (rank 1, diff 0) @ NQ
    # │   │   └── 0 @ OQ
    # │   ├── (rank 1, diff 1) @ SQ
    # │   │   └── 1 @ IQ
    # │   ├── (rank 1, diff 2) @ BQ
    # │   │   └── 2 @ CQ
    # │   ├── (rank 1, diff 3) @ HQ
    # │   │   └── 3 @ FQ
    # │   └── (rank 1, diff 4) @ EQ
    # │       └── 4 @ nQ
    # └── another @ jQ
    #     └── another_leaf @ hQ

    assert str(
        anytree.RenderTree(
            hstrat.SampleAncestralRollbacksTriePostprocessor(seed=1)(
                root,
                p_differentia_collision=0.5,
                mutate=False,
            )
        ).by_attr("taxon_label")
    ) == str(
        anytree.RenderTree(
            hstrat.SampleAncestralRollbacksTriePostprocessor(seed=1)(
                root,
                p_differentia_collision=0.5,
                mutate=False,
            )
        ).by_attr("taxon_label")
    )

    assert str(
        anytree.RenderTree(
            hstrat.SampleAncestralRollbacksTriePostprocessor(seed=1)(
                root,
                p_differentia_collision=0.5,
                mutate=False,
            )
        ).by_attr("taxon_label")
    ) != str(
        anytree.RenderTree(
            hstrat.SampleAncestralRollbacksTriePostprocessor(seed=3)(
                root,
                p_differentia_collision=0.5,
                mutate=False,
            )
        ).by_attr("taxon_label")
    )


def test_sample_ancestral_rollbacks_multiple_levels_mutate():
    root = impl.TrieInnerNode(rank=None, differentia=None)
    impl.TrieInnerNode(rank=0, differentia=101, parent=root)
    for i in range(5):
        node = impl.TrieInnerNode(
            rank=1, differentia=i, parent=root.children[0]
        )
        impl.TrieLeafNode(parent=node, taxon_label=f"{i}")
    another = impl.TrieLeafNode(parent=root, taxon_label="another")
    impl.TrieLeafNode(parent=another, taxon_label="another_leaf")

    # (rank None, diff None) @ KQ
    # ├── (rank 0, diff 101) @ MQ
    # │   ├── (rank 1, diff 0) @ NQ
    # │   │   └── 0 @ OQ
    # │   ├── (rank 1, diff 1) @ SQ
    # │   │   └── 1 @ IQ
    # │   ├── (rank 1, diff 2) @ BQ
    # │   │   └── 2 @ CQ
    # │   ├── (rank 1, diff 3) @ HQ
    # │   │   └── 3 @ FQ
    # │   └── (rank 1, diff 4) @ EQ
    # │       └── 4 @ nQ
    # └── another @ jQ
    #     └── another_leaf @ hQ

    before = str(anytree.RenderTree(root).by_attr("taxon_label"))
    res = hstrat.SampleAncestralRollbacksTriePostprocessor(seed=1)(
        root,
        p_differentia_collision=0.5,
        mutate=False,
    )
    after = str(anytree.RenderTree(root).by_attr("taxon_label"))

    assert before == after
    assert str(anytree.RenderTree(res).by_attr("taxon_label")) != after



from copy import deepcopy
import pickle
import tempfile
import unittest

from hstrat import hstrat


class TestHereditaryStratum(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_deposition_rank(self):
        assert (
            hstrat.HereditaryStratum(
                deposition_rank=42,
            ).GetDepositionRank()
            == 42
        )

    def test_differentia_generation(self):
        original1 = hstrat.HereditaryStratum(
            deposition_rank=42,
        )
        copy1 = deepcopy(original1)
        original2 = hstrat.HereditaryStratum(
            deposition_rank=42,
        )

        assert original1 == copy1
        assert original1 != original2
        assert copy1 != original2

        assert original1.GetDifferentia() == copy1.GetDifferentia()
        assert original1.GetDifferentia() != original2.GetDifferentia()
        assert copy1.GetDifferentia() != original2.GetDifferentia()

    def test_equality1(self):
        assert hstrat.HereditaryStratum() != hstrat.HereditaryStratum()
        stratum1 = hstrat.HereditaryStratum()
        stratum2 = stratum1
        assert stratum1 == stratum2
        assert stratum1 == deepcopy(stratum2)

    def test_equality2(self):
        def stratum_factory():
            return hstrat.HereditaryStratum(deposition_rank=42)

        assert stratum_factory() != stratum_factory()
        stratum1 = stratum_factory()
        stratum2 = stratum1
        assert stratum1 == stratum2
        assert stratum1 == deepcopy(stratum2)

    def test_pickle(self):
        original = hstrat.HereditaryStratum()
        with tempfile.TemporaryDirectory() as tmp_path:
            with open(f"{tmp_path}/data", "wb") as tmp_file:
                pickle.dump(original, tmp_file)

            with open(f"{tmp_path}/data", "rb") as tmp_file:
                reconstituted = pickle.load(tmp_file)
                assert reconstituted == original
                assert reconstituted != hstrat.HereditaryStratum()


if __name__ == "__main__":
    unittest.main()



import itertools as it

import numpy as np
import pytest

from hstrat._auxiliary_lib import all_same
from hstrat.hstrat import geom_seq_nth_root_tapered_algo


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        pytest.param(1, marks=pytest.mark.heavy_3a),
        2,
        3,
        7,
        9,
        pytest.param(42, marks=pytest.mark.heavy_2a),
        pytest.param(97, marks=pytest.mark.heavy_2a),
        pytest.param(100, marks=pytest.mark.heavy_2a),
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        pytest.param(1, marks=pytest.mark.heavy_3b),
        2,
        5,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        pytest.param(
            it.chain(
                range(10**3),
                np.logspace(10, 32, num=50, base=2, dtype="int"),
            ),
            marks=pytest.mark.heavy_3c,
        ),
        (i for i in range(10) for __ in range(2)),
        (10 - i for i in range(10) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**2,
            size=10,
        ),
        pytest.param(
            np.random.default_rng(1).integers(
                low=0,
                high=2**16,
                size=10,
            ),
            marks=pytest.mark.heavy_2b,
        ),
    ],
)
def test_impl_consistency(degree, interspersal, time_sequence):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    impls = [*geom_seq_nth_root_tapered_algo._GenDropRanks_impls]
    instances = [impl(spec) for impl in impls]
    for num_strata_deposited in time_sequence:
        assert all_same(
            it.chain(
                (
                    sorted(
                        impl(spec)(
                            policy,
                            num_strata_deposited,
                            policy.IterRetainedRanks(num_strata_deposited),
                        )
                    )
                    for impl in impls
                ),
                (
                    sorted(
                        instance(
                            policy,
                            num_strata_deposited,
                            policy.IterRetainedRanks(num_strata_deposited),
                        )
                    )
                    for instance in instances
                ),
            )
        )


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        pytest.param(97, marks=pytest.mark.heavy_3a),
        pytest.param(100, marks=pytest.mark.heavy_3a),
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        pytest.param(1, marks=pytest.mark.heavy_3b),
        2,
        5,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        pytest.param(range(10**4), marks=pytest.mark.heavy_3c),
        (i for i in range(10**2) for __ in range(2)),
        pytest.param(
            np.random.default_rng(1).integers(
                low=0,
                high=2**32,
                size=10**2,
            ),
            marks=pytest.mark.heavy_3c,
        ),
    ],
)
def test_policy_consistency(degree, interspersal, time_sequence):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_tapered_algo.GenDropRanks(spec)
    for num_strata_deposited in time_sequence:
        policy_requirement = {
            *policy.IterRetainedRanks(
                num_strata_deposited,
            )
        } - {
            *policy.IterRetainedRanks(
                num_strata_deposited + 1,
            )
        }
        for which in (
            instance,
            geom_seq_nth_root_tapered_algo.GenDropRanks(spec),
        ):
            assert sorted(
                which(
                    policy,
                    num_strata_deposited,
                    policy.IterRetainedRanks(num_strata_deposited),
                )
            ) == sorted(policy_requirement)


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_eq(degree, interspersal):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_tapered_algo.GenDropRanks(spec)

    assert instance == instance
    assert instance == geom_seq_nth_root_tapered_algo.GenDropRanks(spec)
    assert instance is not None



import typing

import matplotlib as mpl
from matplotlib import pyplot as plt
import opytional as opyt

from ..._auxiliary_lib import scale_luminosity
from ...genome_instrumentation._HereditaryStratigraphicColumn import (
    HereditaryStratigraphicColumn,
)


def stratum_retention_dripplot(
    stratum_retention_policy: typing.Any,
    num_generations: int,
    do_show: bool = False,
    ax: typing.Optional[plt.matplotlib.axes.Axes] = None,
    draw_extant_history: bool = True,
    draw_extinct_history: bool = True,
    draw_extinct_placeholders: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
) -> plt.matplotlib.axes.Axes:
    """Show history of retained and purged strata at a particular generation.

    Plots position of retained strata within a hereditary stratigraphic
    column over successive depositions under a particular stratum retention
    policy.

    Parameters
    ----------
    stratum_retention_policy : any
        Object specifying stratum retention policy.
    num_generations : int
        Number of generations to plot.
    do_show : bool, optional
        Whether to show() the plot automatically.
    ax : matplotlib/pylab axes, optional
        If a valid matplotlib.axes.Axes instance, the plot is drawn in that
        Axes. By default (None), a new axes is created.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
    elif not isinstance(ax, plt.matplotlib.axes.Axes):
        raise ValueError(f"Invalid argument for ax: {ax}")

    # instantiate column if IterRetainedRanks not available
    column_t = typing.Optional[HereditaryStratigraphicColumn]
    column: column_t = opyt.apply_if_or_else(
        stratum_retention_policy.IterRetainedRanks,
        lambda __: None,
        lambda: HereditaryStratigraphicColumn(
            stratum_retention_policy=stratum_retention_policy,
        ),
    )

    for gen in progress_wrap(range(1, num_generations)):
        for rank in stratum_retention_policy.GenDropRanks(
            gen,
            opyt.apply_if_or_else(
                stratum_retention_policy.IterRetainedRanks,
                lambda x: x(gen),
                lambda: column.IterRetainedRanks(),
            ),
        ):
            if draw_extinct_placeholders:
                ax.plot(
                    rank,
                    num_generations - 1,
                    ms=20 / max(0.2 * num_generations, 1),
                    marker="v",
                    markerfacecolor="None",
                    markeredgecolor="w",
                    markeredgewidth=4 / max(0.2 * num_generations, 1),
                )
                ax.plot(
                    rank,
                    num_generations - 1,
                    ms=max(
                        20 / max(0.2 * num_generations, 1),
                        4,
                    ),
                    marker="v",
                    markerfacecolor=scale_luminosity(
                        "r", max(10 / max(0.2 * num_generations, 1), 1)
                    ),
                    markeredgecolor="r",
                    markeredgewidth=2 / max(0.05 * num_generations, 1),
                )

            if not draw_extinct_history:
                break
            ax.plot(
                rank,
                gen,
                ms=20 / max(0.2 * num_generations, 1),
                marker="v",
                markerfacecolor="None",
                markeredgecolor="w",
                markeredgewidth=4 / max(0.2 * num_generations, 1),
            )
            ax.plot(
                [rank, rank],
                [rank, gen],
                "w",
                lw=4 / max(0.2 * num_generations, 1),
            )
            ax.plot([rank, rank], [rank, gen], "r")
            ax.plot(
                rank,
                gen,
                ms=max(
                    20 / max(0.2 * num_generations, 1),
                    4,
                ),
                marker="v",
                markerfacecolor=scale_luminosity(
                    "r", max(10 / max(0.2 * num_generations, 1), 1)
                ),
                markeredgecolor="r",
                markeredgewidth=2 / max(0.05 * num_generations, 1),
            )
        opyt.apply_if(column, lambda x: x.DepositStratum())

    for remaining_rank in opyt.apply_if_or_else(
        stratum_retention_policy.IterRetainedRanks,
        lambda x: x(num_generations),
        lambda: column.IterRetainedRanks(),
    ):
        ax.plot(
            remaining_rank,
            num_generations - 1,
            ms=20 / max(0.2 * num_generations, 1),
            marker="v",
            markerfacecolor="None",
            markeredgecolor="w",
            markeredgewidth=4 / max(0.2 * num_generations, 1),
        )
        if draw_extant_history:
            ax.plot(
                [remaining_rank, remaining_rank],
                [remaining_rank, num_generations - 1],
                "w",
                lw=4 / max(0.2 * num_generations, 1),
            )
            ax.plot(
                [remaining_rank, remaining_rank],
                [remaining_rank, num_generations - 1],
                "k",
            )
        ax.plot(
            remaining_rank,
            num_generations - 1,
            ms=max(
                20 / max(0.2 * num_generations, 1),
                4,
            ),
            marker="v",
            markerfacecolor="k",
            markeredgecolor="k",
            markeredgewidth=2 / max(0.05 * num_generations, 1),
        )

    # make space for extant underbar
    ymin, ymax = ax.get_ylim()
    ymid = (ymin + ymax) / 2
    ax.set_ylim([min(ymin, ymid - 2), max(ymax, ymid + 2)])
    # make space for triangle markers
    ymin, ymax = ax.get_ylim()
    ax.set_ylim([ymin, ymax + 2])
    ax.xaxis.set_major_locator(
        mpl.ticker.MaxNLocator(
            nbins="auto",
            steps=[1, 2, 5, 10],
            integer=True,
            min_n_ticks=0,
        )
    )
    ax.yaxis.set_major_locator(
        mpl.ticker.MaxNLocator(
            nbins="auto",
            steps=[1, 2, 5, 10],
            integer=True,
            min_n_ticks=0,
        )
    )

    # strip any negative xticks
    ax.set_xticks([tick for tick in ax.get_xticks() if tick >= 0])

    # make space for rectangle
    xmin, xmax = ax.get_xlim()
    ax.set_xlim(
        [
            min(xmin, -1),
            max(xmax, num_generations),
        ]
    )
    height = min(3, num_generations)
    ax.add_patch(
        mpl.patches.Rectangle(
            xy=(-0.5, num_generations - 1 - height / 2),
            width=num_generations,
            height=height,
            facecolor=scale_luminosity("C0", 1.5),
            edgecolor=scale_luminosity("C0", 0.5),
            linewidth=6 / max(0.2 * num_generations, 1),
        )
    )

    ax.set_xlabel("Position (Rank)")
    ax.set_ylabel("Retention History\n(Generations)")

    if do_show:
        plt.show()

    return ax



import json
import logging

import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import log_once_in_a_row


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_policy_to_records(
    retention_policy,
    caplog,
):
    records = hstrat.policy_to_records(retention_policy)
    assert records == hstrat.policy_to_records(retention_policy)
    for entry in [
        "policy_algo",
        "policy",
        "hstrat_version",
    ]:
        assert entry in records
        assert isinstance(records[entry], str)

    assert "policy_spec" in records
    assert isinstance(records["policy_spec"], dict)


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_policy_to_records_then_from_records(
    retention_policy,
    caplog,
):

    assert hstrat.policy_to_records(
        retention_policy
    ) == hstrat.policy_to_records(retention_policy)
    reconstituted = hstrat.policy_from_records(
        hstrat.policy_to_records(retention_policy),
    )
    assert reconstituted == retention_policy


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_policy_to_records_then_from_records_json(
    retention_policy,
    caplog,
):
    records = hstrat.policy_to_records(retention_policy)
    json_str = json.dumps(records)
    reconstituted = hstrat.policy_from_records(json.loads(json_str))
    assert reconstituted == retention_policy


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_policy_to_records_version_warning(
    retention_policy,
    caplog,
):
    # adapted from https://stackoverflow.com/a/48113200
    with caplog.at_level(logging.INFO):
        # clear caplog
        caplog.clear()

        # create a record with a phony version
        record = hstrat.policy_to_records(retention_policy)
        record["hstrat_version"] = "0.0.0"

        # reconstruct phony record
        hstrat.policy_from_records(record)

        # it should have logged a warning
        assert len(caplog.records) >= 1

        # log something to keep the lru cache from supressing the above calls
        log_once_in_a_row("foo")



from collections import defaultdict

import more_itertools as mit
import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_mark_origin_time_delta_asexual import (
    alifestd_mark_origin_time_delta_asexual,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._alifestd_unfurl_traversal_postorder_asexual import (
    alifestd_unfurl_traversal_postorder_asexual,
)

_UNSAFE_SYMBOLS = (";", "(", ")", ",", "[", "]", ":", "'")


def alifestd_as_newick_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    label_key: str = "id",
    progress_wrap=lambda x: x,
) -> str:
    """Convert phylogeny dataframe to Newick format."""

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    if "origin_time_delta" in phylogeny_df.columns:
        pass
    elif "origin_time" in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_origin_time_delta_asexual(
            phylogeny_df, mutate=True
        )
    else:
        phylogeny_df["origin_time_delta"] = np.nan

    # adapted from https://github.com/niemasd/TreeSwift/blob/63b8979fb5e616ba89079d44e594682683c1365e/treeswift/Node.py#L129
    child_reps = defaultdict(list)
    postorder_ids = alifestd_unfurl_traversal_postorder_asexual(phylogeny_df)
    for id_ in progress_wrap(postorder_ids):

        if label_key is None:
            label = ""
        else:
            label = (
                id_ if label_key == "id" else phylogeny_df.at[id_, label_key]
            )
            label = f"{label}"
            if any(c in label for c in _UNSAFE_SYMBOLS):
                label = f"'{label}'"

        origin_time_delta = phylogeny_df.at[id_, "origin_time_delta"]

        if not np.isnan(origin_time_delta):
            label = f"{label}:{origin_time_delta}"
            if "." in label:
                label = label.rstrip("0").rstrip(".")

        children = child_reps.pop(id_, tuple())
        if children:
            label = f"({','.join(children)}){label}"

        ancestor_id = phylogeny_df.at[id_, "ancestor_id"]
        child_reps[ancestor_id].append(label)

    return ";\n".join(map(mit.one, child_reps.values())) + ";"



import networkx as nx
import pytest

from hstrat import hstrat


def test_empty():
    tree = nx.DiGraph()
    assert (
        hstrat.descend_template_phylogeny_networkx(
            tree, hstrat.HereditaryStratigraphicColumn()
        )
        == []
    )


def test_singleton():
    tree = nx.DiGraph()
    tree.add_nodes_from([0])

    ancestor = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(19)
    assert hstrat.descend_template_phylogeny_networkx(tree, ancestor) == [
        ancestor
    ]


def test_tree():
    tree = nx.DiGraph()
    tree.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4)])
    seed_column = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(1)
    result = hstrat.descend_template_phylogeny_networkx(tree, seed_column)
    assert len(result) == 2
    assert [*tree.nodes].index(2) < [*tree.nodes].index(4)

    col2, col4 = result

    assert col2.GetNumStrataDeposited() == 4
    assert col4.GetNumStrataDeposited() == 5
    assert (
        hstrat.estimate_patristic_distance_between(
            col2, col4, "maximum_likelihood", "arbitrary"
        )
        == 3
    )


def test_forest():
    tree = nx.DiGraph()
    tree.add_edges_from([(0, 1), (0, 2), (2, 3)])
    tree.add_edges_from([(4, 5), (4, 6), (6, 7)])
    with pytest.raises(ValueError):
        hstrat.descend_template_phylogeny_networkx(
            tree, hstrat.HereditaryStratigraphicColumn()
        )


def test_tree_internal_extant_nodes():
    tree = nx.DiGraph()
    tree.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4)])
    seed_column = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(1)
    result = hstrat.descend_template_phylogeny_networkx(
        tree, seed_column, extant_nodes=[2, 3, 4]
    )
    assert len(result) == 3

    col2, col3, col4 = result

    assert col2.GetNumStrataDeposited() == 4
    assert col3.GetNumStrataDeposited() == 4
    assert col4.GetNumStrataDeposited() == 5
    assert (
        hstrat.estimate_patristic_distance_between(
            col2, col4, "maximum_likelihood", "arbitrary"
        )
        == 3
    )
    assert (
        hstrat.estimate_patristic_distance_between(
            col3, col4, "maximum_likelihood", "arbitrary"
        )
        == 1
    )
    assert (
        hstrat.estimate_patristic_distance_between(
            col2, col3, "maximum_likelihood", "arbitrary"
        )
        == 2
    )



import typing

import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ..priors._detail import PriorBase
from ._estimate_rank_of_mrca_between import estimate_rank_of_mrca_between


def estimate_patristic_distance_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    estimator: str,
    prior: typing.Union[typing.Literal["arbitrary", "uniform"], PriorBase],
) -> typing.Optional[float]:
    """Estimate the total phylogenetic distance along the branch path connecting
    the columns.

    Branch length here is in terms of number of generations elapsed. So the
    calculated distance estimates the sum of the number of generations elapsed
    from each to their most recent common ancestor.

    Parameters
    ----------
    estimator : {"maximum_likelihood", "unbiased"}
        What estimation method should be used? Options are "maximum_likelihood"
        or "unbiased".

        See `estimate_ranks_since_mrca_with` for discussion of estimator
        options.
    prior : {"arbitrary", "uniform"} or object implementing prior interface
        Prior probability density distribution over possible generations of the
        MRCA.

        See `estimate_rank_of_mrca_between` for discussion of prior
        options.

    Returns
    -------
    float, optional
        Estimate of patristic distance, unless first and second definitively
        share no common ancestor in which case None will be returned.

    See Also
    --------
    calc_patristic_distance_bounds_between :
        Calculates confidence intervals for patristic distance between two
        hereditary stratigraphic columns.
    does_definitively_have_no_common_anestor :
        Does the hereditary stratigraphic record definitively prove that first and second could not possibly share a common ancestor?
    """

    est_rank_of_mrca_between = estimate_rank_of_mrca_between(
        first,
        second,
        estimator=estimator,
        prior=prior,
    )
    max_patristic_distance = (
        first.GetNumStrataDeposited() - 1 + second.GetNumStrataDeposited() - 1
    )

    return opyt.apply_if(
        est_rank_of_mrca_between,
        lambda est: max_patristic_distance - 2 * est,
    )



from ._ArbitraryPrior import ArbitraryPrior
from ._ExponentialPrior import ExponentialPrior
from ._GeometricPrior import GeometricPrior
from ._UniformPrior import UniformPrior

__all__ = [
    "ArbitraryPrior",
    "GeometricPrior",
    "ExponentialPrior",
    "UniformPrior",
]



import typing

import dendropy as dp
import opytional as opyt

from ..._auxiliary_lib import cast_int_lossless
from ...genome_instrumentation import HereditaryStratigraphicColumn
from ._descend_template_phylogeny import descend_template_phylogeny


def descend_template_phylogeny_dendropy(
    tree: dp.Tree,
    seed_column: HereditaryStratigraphicColumn,
    extant_nodes: typing.Optional[typing.Iterable[dp.Node]] = None,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicColumn]:
    """Generate a population of hereditary stratigraphic columns that could
    have resulted from the template phylogeny.

    Parameters
    ----------
    tree : dendropy.Tree
        Phylogeny record as a dendropy Tree.
    seed_column : HereditaryStratigraphicColumn
        Hereditary stratigraphic column to seed at root node of phylogeny.

        Returned hereditary stratigraphic column population will be generated
        as if repeatedly calling `CloneDescendant()` on `seed_column`. As such,
        specifies configuration (i.e., differentia bit width and stratum
        retention policy) for returned columns. May already have strata
        deposited, which will be incorporated into generated extant population.
    extant_nodes : optional list of dendropy.Node
        Which organisms should hereditary stratigraphic columns be created for?

        Designates content and order of returned list of hereditary
        stratigraphic column.

        If None, hereditary stratigraphic columns will be created for all
        phylogenetic leaves (organisms without offspring) in order of
        appearance in `tree.leaf_node_iter()`.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    list of HereditaryStratigraphicColumn
        Population of hereditary stratigraphic columns for extant lineage
        members (i.e., phylogeny leaf nodes).

        Columns ordered in order of appearance of corresponding extant organism
        id.
    """

    return descend_template_phylogeny(
        ascending_lineage_iterables=(
            extant_node.ancestor_iter(
                inclusive=True,
            )
            for extant_node in opyt.or_value(
                extant_nodes, tree.leaf_node_iter()
            )
        ),
        descending_tree_iterable=tree.levelorder_node_iter(),
        get_parent=lambda node: node.parent_node,
        get_stem_length=lambda node: cast_int_lossless(
            node.edge_length, action="warn", context="edge length"
        ),
        seed_column=seed_column,
        demark=id,
        progress_wrap=progress_wrap,
    )



import typing

import numpy as np

from ..._auxiliary_lib import (
    iter_monotonic_equivalencies,
    jit,
    jit_numba_uint8_arr_t,
    jit_numba_uint16_arr_t,
    jit_numba_uint32_arr_t,
    jit_numba_uint64_arr_t,
)
from ...frozen_instrumentation import HereditaryStratigraphicSpecimen


@jit(
    [
        (
            jit_numba_uint64_arr_t,
            differentia_array_t,
            jit_numba_uint64_arr_t,
            differentia_array_t,
        )
        for differentia_array_t in (
            jit_numba_uint8_arr_t,
            jit_numba_uint16_arr_t,
            jit_numba_uint32_arr_t,
            jit_numba_uint64_arr_t,
        )
    ],
    nopython=True,
)
def _compare_differentia_at_common_ranks(
    first_ranks: np.ndarray,
    first_differentiae: np.ndarray,
    second_ranks: np.ndarray,
    second_differentiae: np.ndarray,
) -> typing.Tuple[int, bool]:
    for pos1, pos2 in iter_monotonic_equivalencies(first_ranks, second_ranks):
        assert first_ranks[pos1] == second_ranks[pos2]
        yield (
            # must convert to Python int; although integral,
            # numpy ints are experiencing unwanted conversion to floats
            # note that specified type signature guarantees that arrays
            # are not inadvertently of float type
            int(first_ranks[pos1]),
            first_differentiae[pos1] == second_differentiae[pos2],
        )


def iter_mutual_ranks(
    first: HereditaryStratigraphicSpecimen,
    second: HereditaryStratigraphicSpecimen,
    compare: bool = False,
) -> typing.Union[
    typing.Iterator[int],
    typing.Iterator[typing.Tuple[int, bool]],
]:
    """Iterate over ranks with matching strata between columns in ascending
    order."""
    if compare:
        return _compare_differentia_at_common_ranks(
            first.GetRankIndex(),
            first.GetDifferentiaVals(),
            second.GetRankIndex(),
            second.GetDifferentiaVals(),
        )

    else:
        assert all(
            first.GetRankIndex()[first_idx] == second.GetRankIndex()[second]
            and isinstance(first.GetRankIndex()[first_idx], np.integer)
            for first_idx, second_idx in iter_monotonic_equivalencies(
                first.GetRankIndex(), second.GetRankIndex()
            )
        )
        return (
            # convert to Python int because elsewhere
            # numpy ints are experiencing unwanted conversion to floats
            int(first.GetRankIndex()[first_idx])
            for first_idx, _second_idx in iter_monotonic_equivalencies(
                first.GetRankIndex(), second.GetRankIndex()
            )
        )



import itertools as it

import numpy as np
import pytest

from hstrat.hstrat import geom_seq_nth_root_algo


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        pytest.param(1, marks=pytest.mark.heavy_3a),
        2,
        3,
        7,
        9,
        pytest.param(42, marks=pytest.mark.heavy_2a),
        pytest.param(97, marks=pytest.mark.heavy_2a),
        pytest.param(100, marks=pytest.mark.heavy_2a),
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        pytest.param(1, marks=pytest.mark.heavy_3b),
        2,
        5,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        pytest.param(
            it.chain(
                range(10**3),
                np.logspace(10, 32, num=50, base=2, dtype="int"),
            ),
            marks=pytest.mark.heavy_3c,
        ),
        (i for i in range(10) for __ in range(2)),
        (10 - i for i in range(10) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**2,
            size=10,
        ),
        pytest.param(
            np.random.default_rng(1).integers(
                low=0,
                high=2**16,
                size=10,
            ),
            marks=pytest.mark.heavy_2b,
        ),
    ],
)
def test_policy_consistency(degree, interspersal, time_sequence):
    policy = geom_seq_nth_root_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_algo.CalcMrcaUncertaintyRelUpperBound(
        spec,
    )
    for num_strata_deposited in time_sequence:
        for actual_mrca_rank in (
            np.random.default_rng(num_strata_deposited,).integers(
                low=0,
                high=num_strata_deposited,
                size=10**2,
            )
            if num_strata_deposited
            else iter(())
        ):
            policy_requirement = policy.CalcMrcaUncertaintyRelExact(
                num_strata_deposited,
                num_strata_deposited,
                actual_mrca_rank,
            )
            for which in (
                instance,
                geom_seq_nth_root_algo.CalcMrcaUncertaintyRelUpperBound(spec),
            ):
                assert (
                    which(
                        policy,
                        num_strata_deposited,
                        num_strata_deposited,
                        actual_mrca_rank,
                    )
                    >= policy_requirement
                )


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_eq(degree, interspersal):
    policy = geom_seq_nth_root_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_algo.CalcMrcaUncertaintyRelUpperBound(spec)

    assert instance == instance
    assert instance == geom_seq_nth_root_algo.CalcMrcaUncertaintyRelUpperBound(
        spec,
    )
    assert instance is not None


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_negative_index(degree, interspersal):
    policy = geom_seq_nth_root_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_algo.CalcMrcaUncertaintyRelUpperBound(spec)

    for diff in range(1, 100):
        assert instance(policy, 100, 100, -diff,) == instance(
            policy,
            100,
            100,
            99 - diff,
        )

        assert instance(policy, 101, 100, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 150, 100, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 101, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 150, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )



def get_package_name(qual_name: str) -> str:
    """
    Gets the package name given the fully qualified module name.

    Examples
    --------
    >>> get_package_name("baz")
    ''
    >>> get_package_name("foo.bar")
    'foo'

    Parameters
    ----------
    qual_name : str
        The fully qualified module name, which usually should be `__name__`
    """
    *package_names, _module_name = qual_name.split(".")
    return ".".join(package_names)



import itertools as it
import unittest

from hstrat._auxiliary_lib import give_len


class TestGiveLen(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_finite_iterator(self):
        x = give_len(iter(range(10)), 10)
        assert len(x) == 10
        assert [*x] == [*range(10)]

    def test_finite_iterable(self):
        x = give_len(range(10), 12)
        assert len(x) == 12
        assert [*reversed(x)] == [*reversed(range(10))]

    def test_override(self):
        x = give_len("asdf", 42)
        assert len(x) == 42
        assert [*x] == [*"asdf"]
        assert x == "asdf"
        assert "asdf" == x

    def test_infinite_iter(self):
        x = give_len(it.count(4), 101)
        assert len(x) == 101
        assert next(x) == 4


if __name__ == "__main__":
    unittest.main()



from ._calc_rank_cutoff import calc_rank_cutoff
from ._calc_rank_sep import calc_rank_sep
from ._calc_target_rank import calc_target_rank


def calc_rank_backstop(
    degree: int,
    interspersal: int,
    pow_: int,
    num_strata_deposited: int,
) -> int:
    """What should the most ancient rank retained to cover the `pow`th target
    recency be?

    This rank and all subsequent ranks spaced forward by `_calc_rank_sep`
    positions through recency zero will be retained. Will monotonically
    increase with `num_strata_deposited` and be an even multiple of
    `_calc_rank_sep`.
    """
    rank_cutoff = calc_rank_cutoff(
        degree,
        interspersal,
        pow_,
        num_strata_deposited,
    )
    retained_ranks_sep = calc_rank_sep(
        degree,
        interspersal,
        pow_,
        num_strata_deposited,
    )

    # round UP from rank_cutoff
    # to align evenly with retained_ranks_sep
    # adapted from https://stackoverflow.com/a/14092788
    min_retained_rank = rank_cutoff - (rank_cutoff % -retained_ranks_sep)
    assert min_retained_rank % retained_ranks_sep == 0
    assert min_retained_rank >= rank_cutoff

    # check that even with rounding up, we are still covering the target
    # rank
    # i.e., that the most ancient retained rank (the backstop rank) falls
    # before the target rank so that the target rank is guaranteed within
    # a _calc_rank_sep window
    assert min_retained_rank <= calc_target_rank(
        degree, pow_, num_strata_deposited
    )

    # more sanity checks on range of output value
    if num_strata_deposited == 0:
        assert min_retained_rank == 0
    else:
        assert 0 <= min_retained_rank <= num_strata_deposited - 1

    # backstop rank synonomous w/ the most ancient (min) retained rank
    return min_retained_rank



import typing

import mpmath as mp

from ._calc_common_ratio import calc_common_ratio


def calc_target_recency(
    degree: int,
    pow_: int,
    num_strata_deposited: int,
) -> typing.Union[float, mp.mpf]:
    """ "What should the target recency of the `pow`'th exponentially
    distributed coverage target be when `num_strata_deposited`?

    Will strictly increase with `num_strata_deposited`.
    """
    common_ratio = calc_common_ratio(degree, num_strata_deposited)
    try:
        res = common_ratio**pow_
    except OverflowError:
        res = mp.mpf(common_ratio) ** pow_
    return res



import itertools as it
import os
import typing

import alifedata_phyloinformatics_convert as apc
import dendropy as dp
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import alifestd_validate

from . import _impl as impl

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize("use_impl", ["cpp", "python", None])
def test_empty_population(use_impl: typing.Optional[str]):
    population = []
    tree = hstrat.build_tree_searchtable(population, use_impl=use_impl)

    assert len(tree) == 0
    assert alifestd_validate(tree)


@pytest.mark.parametrize("use_impl", ["cpp", "python", None])
@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/grandchild_and_aunt.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandchild_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandchild.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtriplets_and_aunt.newick"
        ),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtriplets_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandtriplets.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/grandtwins_and_aunt.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtwins_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandtwins.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/justroot.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/triplets.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/twins.newick"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize("wrap", [lambda x: x, hstrat.col_to_specimen])
def test_smoke(
    use_impl: typing.Optional[str],
    orig_tree: dp.Tree,
    retention_policy: object,
    wrap: typing.Callable,
):
    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(10),
    )

    reconst_df = hstrat.build_tree_searchtable(
        [*map(wrap, extant_population)], use_impl=use_impl
    )

    assert alifestd_validate(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    reconst_tree.collapse_unweighted_edges()

    common_namespace = dp.TaxonNamespace()
    orig_tree.migrate_taxon_namespace(common_namespace)
    reconst_tree.migrate_taxon_namespace(common_namespace)

    original_distance_matrix = orig_tree.phylogenetic_distance_matrix()
    reconstructed_distance_matrix = reconst_tree.phylogenetic_distance_matrix()

    taxa = [node.taxon for node in orig_tree.leaf_node_iter()]

    for a, b in it.combinations(taxa, 2):
        assert original_distance_matrix.distance(
            a, b
        ) == reconstructed_distance_matrix.distance(a, b)


@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/grandchild_and_aunt.newick"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_bad_use_impl(orig_tree: dp.Tree, retention_policy: object):
    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(10),
    )

    with pytest.raises(ValueError):
        hstrat.build_tree_searchtable(extant_population, use_impl="foobar")



import pickle
import tempfile

from hstrat.hstrat import nominal_resolution_algo


def test_eq():
    spec = nominal_resolution_algo.PolicySpec()
    assert spec == spec
    assert spec == nominal_resolution_algo.PolicySpec()


def test_GetEvalCtor():
    # hstrat. is needed for eval()
    from hstrat import hstrat  # noqa

    spec = nominal_resolution_algo.PolicySpec()
    eval_ctor = spec.GetEvalCtor()
    assert eval_ctor.startswith("hstrat.nominal_resolution_algo.PolicySpec(")
    assert eval_ctor.endswith(")")
    reconstituted = eval(eval_ctor)  # noqa
    assert spec == reconstituted


def test_pickle():
    original = nominal_resolution_algo.PolicySpec()
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


def test_init():
    pass


def test_GetAlgoIdentifier():
    spec = nominal_resolution_algo.PolicySpec()
    assert spec.GetAlgoIdentifier()


def test_GetAlgoTitle():
    spec = nominal_resolution_algo.PolicySpec()
    assert spec.GetAlgoTitle()


def test_repr():
    spec = nominal_resolution_algo.PolicySpec()
    assert spec.GetAlgoIdentifier() in repr(spec)


def test_str():
    spec = nominal_resolution_algo.PolicySpec()
    assert spec.GetAlgoTitle() in str(spec)



def calc_common_ratio(
    degree: int,
    num_strata_deposited: int,
) -> float:
    """What should the base of the exponential distribution of retained
    ranks be?"""
    # base ** degree == num_strata_deposited
    # take the degree'th root of each side...
    return num_strata_deposited ** (1 / degree)



import pytest

from hstrat import hstrat
from hstrat.phylogenetic_inference.priors._BubbleWrappedPrior import (
    BubbleWrappedPrior,
)
import hstrat.phylogenetic_inference.priors._detail as detail


def test_base_class():
    assert issubclass(BubbleWrappedPrior, detail.PriorBase)


def test_calc_interval_probability_proxy():
    wrapee = hstrat.ArbitraryPrior()
    wrapped = BubbleWrappedPrior(wrapee)
    for params in (0, 50), (7, 8), (99, 100), (78, 100):
        assert wrapped.CalcIntervalProbabilityProxy(
            *params
        ) == wrapee.CalcIntervalProbabilityProxy(*params)

    with pytest.raises(AssertionError):
        wrapped.CalcIntervalProbabilityProxy(7, 7)
        wrapped.CalcIntervalProbabilityProxy(42, 41)
        wrapped.CalcIntervalProbabilityProxy(-1, 41)


def test_calc_interval_conditioned_mean():
    wrapee = hstrat.ArbitraryPrior()
    wrapped = BubbleWrappedPrior(wrapee)
    for params in (0, 50), (7, 8), (99, 100), (78, 100):
        assert wrapped.CalcIntervalConditionedMean(
            *params
        ) == wrapee.CalcIntervalConditionedMean(*params)

    with pytest.raises(AssertionError):
        wrapped.CalcIntervalConditionedMean(7, 7)
        wrapped.CalcIntervalConditionedMean(42, 41)
        wrapped.CalcIntervalConditionedMean(-1, 41)


def test_sample_interval_conditioned_value():
    wrapee = hstrat.ArbitraryPrior()
    wrapped = BubbleWrappedPrior(wrapee)
    with pytest.raises(NotImplementedError):
        wrapped.SampleIntervalConditionedValue(0, 100)



import re


def collapse_nonleading_whitespace(text: str) -> str:
    """Collapse segments of whitespace not beginning a line to a single
    space."""
    return re.sub(r"(?<=\S)[ \t]+", " ", text, flags=re.MULTILINE)



from estimators import (
    estimate_rank_of_mrca_maximum_likelihood,
    estimate_rank_of_mrca_naive,
    estimate_rank_of_mrca_unbiased,
)
from pairwise import (
    ballpark_patristic_distance_between,
    ballpark_rank_of_mrca_between,
    ballpark_ranks_since_mrca_with,
    calc_patristic_distance_bounds_between,
    calc_rank_of_earliest_detectable_mrca_between,
    calc_rank_of_mrca_bounds_between,
    calc_rank_of_mrca_bounds_provided_confidence_level,
    calc_rank_of_mrca_uncertainty_between,
    calc_ranks_since_earliest_detectable_mrca_with,
    calc_ranks_since_mrca_bounds_provided_confidence_level,
    calc_ranks_since_mrca_bounds_with,
    calc_ranks_since_mrca_uncertainty_with,
    does_definitively_have_no_common_ancestor,
    does_have_any_common_ancestor,
    estimate_patristic_distance_between,
    estimate_rank_of_mrca_between,
    estimate_ranks_since_mrca_with,
)
from population import (
    build_distance_matrix_biopython,
    build_distance_matrix_numpy,
    calc_rank_of_earliest_detectable_mrca_among,
    calc_rank_of_mrca_bounds_among,
    calc_rank_of_mrca_uncertainty_among,
    does_definitively_share_no_common_ancestor,
    does_share_any_common_ancestor,
)
from priors import (
    ArbitraryPrior,
    ExponentialPrior,
    GeometricPrior,
    UniformPrior,
)
from tree import (
    AssignDestructionTimeYoungestPlusOneTriePostprocessor,
    AssignOriginTimeExpectedValueTriePostprocessor,
    AssignOriginTimeNaiveTriePostprocessor,
    AssignOriginTimeNodeRankTriePostprocessor,
    AssignOriginTimeSampleNaiveTriePostprocessor,
    CompoundTriePostprocessor,
    PeelBackConjoinedLeavesTriePostprocessor,
    SampleAncestralRollbacksTriePostprocessor,
    build_tree,
    build_tree_nj,
    build_tree_searchtable,
    build_tree_trie,
    build_tree_trie_ensemble,
    build_tree_upgma,
    trie_postprocess,
)

from . import estimators, pairwise, population, priors, tree

__all__ = [
    "estimators",
    "pairwise",
    "population",
    "priors",
    "tree",
    "estimate_rank_of_mrca_maximum_likelihood",
    "estimate_rank_of_mrca_naive",
    "estimate_rank_of_mrca_unbiased",
    "ballpark_patristic_distance_between",
    "ballpark_rank_of_mrca_between",
    "ballpark_ranks_since_mrca_with",
    "calc_patristic_distance_bounds_between",
    "calc_rank_of_earliest_detectable_mrca_between",
    "calc_rank_of_mrca_bounds_between",
    "calc_rank_of_mrca_bounds_provided_confidence_level",
    "calc_rank_of_mrca_uncertainty_between",
    "calc_ranks_since_earliest_detectable_mrca_with",
    "calc_ranks_since_mrca_bounds_with",
    "calc_ranks_since_mrca_bounds_provided_confidence_level",
    "calc_ranks_since_mrca_uncertainty_with",
    "does_definitively_have_no_common_ancestor",
    "does_have_any_common_ancestor",
    "estimate_patristic_distance_between",
    "estimate_rank_of_mrca_between",
    "estimate_ranks_since_mrca_with",
    "build_distance_matrix_biopython",
    "build_distance_matrix_numpy",
    "calc_rank_of_earliest_detectable_mrca_among",
    "calc_rank_of_mrca_bounds_among",
    "calc_rank_of_mrca_uncertainty_among",
    "does_definitively_share_no_common_ancestor",
    "does_share_any_common_ancestor",
    "ArbitraryPrior",
    "GeometricPrior",
    "ExponentialPrior",
    "UniformPrior",
    "build_tree_nj",
    "build_tree_searchtable",
    "build_tree_trie",
    "build_tree_trie_ensemble",
    "build_tree_upgma",
    "build_tree",
    "trie_postprocess",
    "AssignDestructionTimeYoungestPlusOneTriePostprocessor",
    "AssignOriginTimeExpectedValueTriePostprocessor",
    "AssignOriginTimeNaiveTriePostprocessor",
    "AssignOriginTimeNodeRankTriePostprocessor",
    "AssignOriginTimeSampleNaiveTriePostprocessor",
    "CompoundTriePostprocessor",
    "PeelBackConjoinedLeavesTriePostprocessor",
    "SampleAncestralRollbacksTriePostprocessor",
]



import itertools as it
import typing

import numpy as np
import statsmodels.api as sm


def estimate_binomial_p(
    sampler: typing.Callable,
    confidence: float = 0.99,
    precision: float = 0.01,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.Tuple[float, typing.Tuple[float, float, int]]:
    """Estimate the probability of success in a binomial distribution, along
    with a confidence interval.

    This function uses a given sampler to generate observations, estimating the
    probability of success. It continues sampling until the width of the
    confidence interval is less than or equal to the specified precision.

    Parameters
    ----------
    sampler : typing.Callable
        A no-argument function that returns a boolean indicating success (True)
        or failure (False) for each trial.
    confidence : float, optional
        The confidence level for the confidence interval of the estimated
        probability, by default 0.99.
    precision : float, optional
        The desired precision of the confidence interval (the maximum
        acceptable width of the interval), by default 0.01.

    Returns
    -------
    typing.Tuple[float, typing.Tuple[float, float], int]
        A tuple containing the estimated probability of success, the
        confidence interval (lower bound, upper bound), and the number of
        observations taken.

    Notes
    -----
    The estimation uses the Wilson score interval for calculating the
    confidence interval of the proportion.

    Examples
    --------
    >>> def coin_flip_sampler():
    ...     return np.random.choice([True, False])
    ...
    >>> p_estimate, conf_interval = estimate_binomial_p(coin_flip_sampler)
    """
    num_success = 0
    for num_observed in progress_wrap(it.count(start=1)):
        num_success += bool(sampler())
        confint = sm.stats.proportion_confint(
            count=num_success,
            nobs=num_observed,
            alpha=1 - confidence,
            method="wilson",
        )
        uncertainty = np.ptp(confint)
        if uncertainty <= precision:
            break

    return num_success / num_observed, confint, num_observed



import typing

from ...frozen_instrumentation import HereditaryStratigraphicSpecimen
from ._calc_rank_of_parity_segue_between import (
    calc_rank_of_parity_segue_between,
)


def calc_rank_of_first_retained_disparity_between(
    first: HereditaryStratigraphicSpecimen,
    second: HereditaryStratigraphicSpecimen,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find first mismatching strata between columns.

    Implementation detail. Provides optimized implementation for special
    case where both self and second use the perfect resolution stratum
    retention policy.
    """
    return calc_rank_of_parity_segue_between(
        first,
        second,
        None,
        confidence_level,
    )[1]



import pickle
import tempfile

import pytest

from hstrat.hstrat import geom_seq_nth_root_algo


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_eq(degree, interspersal):
    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert spec == spec
    assert spec == geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert not spec == geom_seq_nth_root_algo.PolicySpec(
        degree, interspersal + 1
    )
    assert not spec == geom_seq_nth_root_algo.PolicySpec(
        degree + 1, interspersal
    )
    assert not spec == geom_seq_nth_root_algo.PolicySpec(
        degree + 1, interspersal + 1
    )


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_GetEvalCtor(degree, interspersal):
    # hstrat. is needed for eval()
    from hstrat import hstrat  # noqa

    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    eval_ctor = spec.GetEvalCtor()
    assert eval_ctor.startswith("hstrat.geom_seq_nth_root_algo.PolicySpec(")
    assert eval_ctor.endswith(")")
    reconstituted = eval(eval_ctor)  # noqa
    assert spec == reconstituted


@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_pickle(degree, interspersal):
    original = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_GetDegree(degree, interspersal):
    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert spec.GetDegree() == degree


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_GetIterspersal(degree, interspersal):
    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert spec.GetInterspersal() == interspersal


def test_GetAlgoIdentifier():
    degree = 1
    interspersal = 2
    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert spec.GetAlgoIdentifier()


def test_GetAlgoTitle():
    degree = 1
    interspersal = 2
    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert spec.GetAlgoTitle()


def test_repr():
    degree = 1
    interspersal = 2
    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert str(degree) in repr(spec)
    assert str(interspersal) in repr(spec)
    assert spec.GetAlgoIdentifier() in repr(spec)


def test_str():
    degree = 1
    interspersal = 2
    spec = geom_seq_nth_root_algo.PolicySpec(degree, interspersal)
    assert str(degree) in str(spec)
    assert str(interspersal) in str(spec)
    assert spec.GetAlgoTitle() in str(spec)



import contextlib
import itertools as it
import random
import typing

import opytional as opyt

from ...._auxiliary_lib import (
    AnyTreeFastPreOrderIter,
    RngStateContext,
    anytree_calc_leaf_counts,
    anytree_has_grandparent,
    anytree_has_sibling,
    anytree_iterative_deepcopy,
    anytree_peel_sibling_to_cousin,
)
from .._impl import TrieInnerNode
from ._detail import TriePostprocessorBase


def _sample_ancestral_rollbacks(
    trie: TrieInnerNode,
    p_differentia_collision: float,
    mutate: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
) -> TrieInnerNode:
    """Implementation detail for `SampleAncestralRollbacks.__call__`.

    See `SampleAncestralRollbacks.__call__` for parameter descriptions.
    """
    if not mutate:
        trie = anytree_iterative_deepcopy(trie, progress_wrap=progress_wrap)

    eligible_nodes = {
        id(node): node
        for node in AnyTreeFastPreOrderIter(trie)
        if anytree_has_sibling(node) and anytree_has_grandparent(node)
    }
    # sequence data structure allows efficient random choice
    possibly_eligible_node_ids = [*eligible_nodes.keys()]

    # number of internal nodes approx equal to the number of branching nodes in
    # a strictly bifurcating/unifurcating tree
    # ... correction for multifurcations (i.e., due to strong selection
    # pressure) should be considered in the future (including whether such
    # corections are necessary in the first place)
    num_leaves = sum(node.is_leaf for node in AnyTreeFastPreOrderIter(trie))
    unzip_opportunities = num_leaves

    if p_differentia_collision <= 0.5:
        # 1 + 1/x + 1/x^2 + ... = x / (x - 1)
        # expected number of successive collisions:
        # 1/x + 1/x^2 + 1/x^3 + ... =  1 / (x - 1)
        # p = 1/x -> x = 1/p
        #  1/x + 1/x^2 + ... = 1 / (1/p - 1)
        # 1/x + 1/x^2 + ... = p / (1 - p)
        # note: does not account for limitations in the number of possible
        # collisions due to tree depth
        collision_succession_corrected_expectation_per_opportunity = (
            p_differentia_collision / (1 - p_differentia_collision)
        )
    else:
        # allow feeding unrealistic p for bounds testing
        collision_succession_corrected_expectation_per_opportunity = (
            p_differentia_collision
        )

    # note:
    # this estimates the expected value of the number of collisions;
    # so some possible outcomes like all max_unzips being performed or no
    # unzips being performed will never occur;
    # an alternate approach would be to sample the number of collisions...
    # the number of collisions might be drawn from a binomial distribution
    # but some care would have to be taken to consider the possibility of
    # successive collisions where the MRCA is rolled back more than one
    # position
    expected_collisions = int(
        collision_succession_corrected_expectation_per_opportunity
        * unzip_opportunities
    )

    def calc_max_unzips() -> bool:
        leaf_counts = anytree_calc_leaf_counts(trie)
        del leaf_counts[id(trie)]  # exclude root
        max_unzips = sum(leaf_counts.values()) - len(
            leaf_counts
        )  # -1 per; last sibling always ineligible for unzip
        return max_unzips

    expected_collisions = min(expected_collisions, calc_max_unzips())

    remaining_collisions = expected_collisions

    progress = iter(progress_wrap(it.count()))
    while remaining_collisions:
        next(progress)
        assert possibly_eligible_node_ids
        target_idx = random.randrange(len(possibly_eligible_node_ids))
        target_id = possibly_eligible_node_ids[target_idx]
        if target_id not in eligible_nodes:
            # swap and pop to update possibly_eligible_node_ids
            # to match eligible_nodes
            (
                possibly_eligible_node_ids[-1],
                possibly_eligible_node_ids[target_idx],
            ) = (
                possibly_eligible_node_ids[target_idx],
                possibly_eligible_node_ids[-1],
            )
            assert possibly_eligible_node_ids[-1] == target_id
            possibly_eligible_node_ids.pop()
            continue

        target_node = eligible_nodes[target_id]

        original_parent = target_node.parent
        grandparent = target_node.parent.parent
        target_node = eligible_nodes[target_id]
        anytree_peel_sibling_to_cousin(target_node)
        remaining_collisions -= 1

        # target node is now ineligible
        # swap and pop out of possibly_eligible_node_ids, eligible_nodes
        (
            possibly_eligible_node_ids[-1],
            possibly_eligible_node_ids[target_idx],
        ) = (
            possibly_eligible_node_ids[target_idx],
            possibly_eligible_node_ids[-1],
        )
        assert possibly_eligible_node_ids[-1] == target_id
        possibly_eligible_node_ids.pop()
        eligible_nodes.pop(target_id)

        # is target's (possibly lone) sibling now ineligible?
        if len(original_parent.children) == 1:
            eligible_nodes.pop(id(original_parent.children[0]))

        for sibling in grandparent.children:
            # peeled off parent is always newly eligible
            # peeled from parent might be newly eligible
            if len(grandparent.children) == 2 or sibling is target_node.parent:
                if anytree_has_grandparent(sibling):
                    assert id(sibling) not in eligible_nodes
                    possibly_eligible_node_ids.append(id(sibling))
                    eligible_nodes[id(sibling)] = sibling
                    assert anytree_has_sibling(sibling)
                    assert anytree_has_grandparent(sibling)

    return trie


class SampleAncestralRollbacksTriePostprocessor(
    TriePostprocessorBase,
):
    """Functor to correct for systematic overestimation of relatedness by
    sampling a compensatory adjustment to trie topology."""

    _seed: typing.Optional[int]

    def __init__(
        self: "SampleAncestralRollbacksTriePostprocessor",
        seed: typing.Optional[int] = None,
    ) -> None:
        """Initialize functor instance.

        Parameters:
        ----------
        seed: int, default
            Controls sampling decisions in the algorithm.

            Pass an int for reproducible output across multiple function calls.
            The default value, 1, ensures reproducible output. Pass None to use
            existing RNG context directly.
        """

        self._seed = seed

    def __call__(
        self: "SampleAncestralRollbacksTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Compensate for bias towards overestimating relatedness due to
        spurious differentia collisions.

        Each rollback operation alters the tree as if a single spurious
        collision had occured; a single branch is adjusted to exhibit the next-
        most-ancient last commonality.

        The number of rollback operations is calculated from the number of
        possible spurious collisions and the probability of spurious collision.
        Unzip targets are sampled randomly using the standard library `random`
        module.

        Parameters:
        ----------
        trie : TrieInnerNode
            The root node of the trie to be unzipped.
        p_differentia_collision : float
            The multiplicative inverse of the number of possible
            differentia.

            This fraction of possible rollbacks are performed.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie.

        Notes:
        ------
        This function assumes underlying shared genesis, so the root node of the
        trie is not eligible for rollback.
        """
        with opyt.apply_if_or_value(
            self._seed,
            RngStateContext,
            contextlib.nullcontext(),
        ):
            return _sample_ancestral_rollbacks(
                trie=trie,
                p_differentia_collision=p_differentia_collision,
                mutate=mutate,
                progress_wrap=progress_wrap,
            )



import typing

from ...genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratumOrderedStoreList,
)
from ._calc_rank_of_last_retained_commonality_between_bsearch import (
    calc_rank_of_last_retained_commonality_between_bsearch,
)
from ._calc_rank_of_last_retained_commonality_between_generic import (
    calc_rank_of_last_retained_commonality_between_generic,
)


def calc_rank_of_last_retained_commonality_between(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    confidence_level: float,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Optional[int]:
    """Find rank of strata commonality before first strata disparity.

    Implementation detail with general-case implementation.
    """

    if (
        first.HasDiscardedStrata()
        or second.HasDiscardedStrata()
        # for performance reasons
        # only apply binary search to stores that support random access
        or not hasattr(first, "_stratum_ordered_store")
        or not isinstance(
            first._stratum_ordered_store, HereditaryStratumOrderedStoreList
        )
        or not hasattr(second, "_stratum_ordered_store")
        or not isinstance(
            second._stratum_ordered_store, HereditaryStratumOrderedStoreList
        )
        # binary search currently requires no spurious collisions
        or first.GetStratumDifferentiaBitWidth() < 64
    ):
        return calc_rank_of_last_retained_commonality_between_generic(
            first,
            second,
            confidence_level=confidence_level,
        )
    else:
        return calc_rank_of_last_retained_commonality_between_bsearch(
            first,
            second,
            confidence_level=confidence_level,
        )



import math

import anytree

from hstrat import hstrat
import hstrat.phylogenetic_inference.tree._impl as impl
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.AssignOriginTimeExpectedValueTriePostprocessor,
        detail.TriePostprocessorBase,
    )


def test_assign_trie_origin_times_expected_value_single_leaf__arbitrary():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.ArbitraryPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert leaf.origin_time == 0
    assert inner.origin_time == 0
    assert root.origin_time == 0


def test_assign_trie_origin_times_expected_value_single_leaf__assigned_property():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.ArbitraryPrior(),
        assigned_property="raspberry",
    )(root, p_differentia_collision=0.25, mutate=True)
    assert leaf.raspberry == 0
    assert inner.raspberry == 0
    assert root.raspberry == 0


def test_assign_trie_origin_times_expected_value_two_leaves__arbitrary():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf1 = impl.TrieLeafNode(parent=inner, taxon_label="A")
    leaf2 = impl.TrieLeafNode(parent=inner, taxon_label="B")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.ArbitraryPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert inner.origin_time == 0
    assert leaf1.origin_time == 0
    assert leaf2.origin_time == 0


def test_assign_trie_origin_times_expected_value1__arbitrary():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=1, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.ArbitraryPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert inner2.origin_time == (1 + 0) / (1 + 0.25)
    assert leaf.origin_time == 1


def test_assign_trie_origin_times_expected_value2__arbitrary():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=2, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.ArbitraryPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert math.isclose(inner1.origin_time, (0.5 + 0 * 0.25) / (1 + 0.25))
    assert math.isclose(
        inner2.origin_time, (2 + inner1.origin_time * 0.25) / (1 + 0.25)
    )
    assert leaf.origin_time == 2


def test_assign_trie_origin_times_expected_value3__arbitrary():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=3, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.ArbitraryPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert math.isclose(inner1.origin_time, (1 + 0 * 0.25) / (1 + 0.25))
    assert math.isclose(
        inner2.origin_time, (3 + inner1.origin_time * 0.25) / (1 + 0.25)
    )
    assert leaf.origin_time == 3


def test_assign_trie_origin_times_expected_value_complex__arbitrary():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)

    inner2 = impl.TrieInnerNode(rank=4, differentia=10, parent=inner1)
    inner3 = impl.TrieInnerNode(rank=7, differentia=7, parent=inner2)

    inner2a = impl.TrieInnerNode(rank=1, differentia=1, parent=inner1)

    leaf3_a = impl.TrieLeafNode(parent=inner3, taxon_label="leaf3a_a")

    impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_a")
    leaf2a_b = impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_b")

    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.ArbitraryPrior()
    )(root, p_differentia_collision=0.25, mutate=True)

    assert root.origin_time == 0
    assert math.isclose(inner1.origin_time, (0 + 0 * 0.25) / 1.25)
    assert math.isclose(
        inner2.origin_time, (5 + inner1.origin_time * 0.25) / 1.25
    )
    assert math.isclose(
        inner3.origin_time, (7 + inner2.origin_time * 0.25) / 1.25
    )
    assert math.isclose(
        inner2a.origin_time, (1 + inner1.origin_time * 0.25) / 1.25
    )
    assert leaf2a_b.origin_time == 1
    assert leaf3_a.origin_time == 7


def test_assign_trie_origin_times_expected_value_single_leaf__uniform():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.UniformPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert leaf.origin_time == 0
    assert inner.origin_time == 0
    assert root.origin_time == 0


def test_assign_trie_origin_times_expected_value_two_leaves__uniform():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf1 = impl.TrieLeafNode(parent=inner, taxon_label="A")
    leaf2 = impl.TrieLeafNode(parent=inner, taxon_label="B")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.UniformPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert inner.origin_time == 0
    assert leaf1.origin_time == 0
    assert leaf2.origin_time == 0


def test_assign_trie_origin_times_expected_value1__uniform():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=1, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.UniformPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert inner1.origin_time == 0
    assert inner2.origin_time == (1 + 0) / (1 + 0.25)
    assert leaf.origin_time == 1


def test_assign_trie_origin_times_expected_value2__uniform():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=2, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.UniformPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert math.isclose(
        inner1.origin_time, (2 * 0.5 + 0 * 0.25) / (2 * 1 + 1 * 0.25)
    )
    assert math.isclose(
        inner2.origin_time, (2 + inner1.origin_time * 0.25) / (1 + 0.25)
    )
    assert leaf.origin_time == 2


def test_assign_trie_origin_times_expected_value3__uniform():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=3, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.UniformPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert math.isclose(
        inner1.origin_time, (3 * 1 + 0 * 0.25) / (3 * 1 + 0.25)
    )
    assert math.isclose(
        inner2.origin_time, (3 + inner1.origin_time * 0.25) / (1 + 0.25)
    )
    assert leaf.origin_time == 3


def test_assign_trie_origin_times_expected_value4__uniform():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=3, differentia=10, parent=inner1)
    inner3 = impl.TrieInnerNode(rank=6, differentia=10, parent=inner2)
    leaf = impl.TrieLeafNode(parent=inner3, taxon_label="A")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.UniformPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert math.isclose(
        inner1.origin_time, (3 * 1 + 0 * 0.25) / (3 * 1 + 0.25)
    )
    expected = (3 * 4 + inner1.origin_time * 0.25) / (3 * 1 + 0.25)
    assert math.isclose(inner2.origin_time, expected)
    assert leaf.origin_time == 6


def test_assign_trie_origin_times_expected_value3__uniform_fork():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=3, differentia=10, parent=inner1)
    inner2b = impl.TrieInnerNode(rank=300, differentia=11, parent=inner1)
    inner3b = impl.TrieInnerNode(rank=309, differentia=12, parent=inner2b)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    leafb = impl.TrieLeafNode(parent=inner3b, taxon_label="X")
    root = hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
        prior=hstrat.UniformPrior()
    )(root, p_differentia_collision=0.25, mutate=True)
    assert root.origin_time == 0
    assert math.isclose(
        inner1.origin_time, (3 * 1 + 0 * 0.25) / (3 * 1 + 0.25)
    )
    assert math.isclose(
        inner2.origin_time, (3 + inner1.origin_time * 0.25) / (1 + 0.25)
    )
    assert leaf.origin_time == 3
    expected = (304 * 9 + inner1.origin_time * 0.25) / (1 * 9 + 0.25)
    assert math.isclose(inner2b.origin_time, expected)
    expected = (309 + inner2b.origin_time * 0.25) / (1 + 0.25)
    assert math.isclose(inner3b.origin_time, expected)
    assert leafb.origin_time == 309


def test_assign_trie_origin_times_naive_mutate():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=1, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root_ = hstrat.AssignOriginTimeNaiveTriePostprocessor(
        prior=hstrat.ArbitraryPrior(),
        assigned_property="raspberry",
    )(root, p_differentia_collision=0.5, mutate=False)
    assert not hasattr(leaf, "raspberry")
    assert not hasattr(inner1, "raspberry")
    assert not hasattr(inner2, "raspberry")
    assert not hasattr(root, "raspberry")

    assert hasattr(root_, "raspberry")
    root = hstrat.AssignOriginTimeNaiveTriePostprocessor(
        prior=hstrat.ArbitraryPrior(),
        assigned_property="raspberry",
    )(root, p_differentia_collision=0.5, mutate=True)

    assert str(anytree.RenderTree(root).by_attr("raspberry")) == str(
        anytree.RenderTree(root_).by_attr("raspberry")
    )



import pandas as pd

from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_try_add_ancestor_list_col_polars import (
    alifestd_try_add_ancestor_list_col_polars,
)
from ._delegate_polars_implementation import (
    DataFrame_T,
    delegate_polars_implementation,
)


@delegate_polars_implementation(alifestd_try_add_ancestor_list_col_polars)
def alifestd_try_add_ancestor_list_col(
    phylogeny_df: DataFrame_T,
    root_ancestor_token: str = "none",
    mutate: bool = False,
) -> DataFrame_T:
    """Add an ancestor_list column to the input DataFrame if the column does
    not already exist.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_make_ancestor_list_col
    """

    assert isinstance(phylogeny_df, pd.DataFrame)
    if "ancestor_id" in phylogeny_df and "ancestor_list" not in phylogeny_df:
        if not mutate:
            phylogeny_df = phylogeny_df.copy()
        phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
            phylogeny_df["id"],
            phylogeny_df["ancestor_id"],
            root_ancestor_token=root_ancestor_token,
        )

    return phylogeny_df



import itertools as it
import typing
import warnings

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ..pairwise import calc_rank_of_mrca_bounds_between
from ._calc_rank_of_earliest_detectable_mrca_among import (
    calc_rank_of_earliest_detectable_mrca_among,
)


def calc_rank_of_mrca_bounds_among(
    population: typing.Iterable[HereditaryStratigraphicArtifact],
    prior: str,
    confidence_level: float = 0.95,
) -> typing.Optional[typing.Tuple[int, int]]:
    """Within what generation range did MRCA fall?

    Calculate bounds on estimate for the number of depositions elapsed
    along the line of descent before the most recent common ancestor among population.

    Parameters
    ----------
    prior : {"arbitrary"}
        Prior probability density distribution over possible generations of the
        MRCA.

        Currently only "arbitrary" supported.
    confidence_level : float, optional
        Bounds must capture what probability of containing the true rank of
        the MRCA? Default 0.95.

    Returns
    -------
    (int, int), optional
        Inclusive lower and then exclusive upper bound on estimate or None
        if no common ancestor between first and second can be resolved with
        sufficient confidence. (Sufficient confidence depends on
        confidence_level.) Also returns None for empty or singleton population.

    Notes
    -----
    Currently implementaiton uses a naive O(n^2) approach. A more efficient
    implementation should be possible.

    The true rank of the MRCA is guaranteed to never fall above the bounds
    but may fall below.
    """
    assert 0.0 <= confidence_level <= 1.0

    pop_tee1, pop_tee2 = it.tee(population)
    if (
        calc_rank_of_earliest_detectable_mrca_among(
            pop_tee1,
            confidence_level=confidence_level,
        )
        is None
    ):
        warnings.warn(
            "Insufficient common ranks between columns to detect common "
            "ancestry at given confidence level."
        )
        return None

    pop_mrca_rank_lower_bound = float("inf")
    pop_mrca_rank_upper_bound = float("inf")
    any_combinations = False
    for (first, second) in it.combinations(pop_tee2, 2):
        any_combinations = True
        assert (
            first.GetStratumDifferentiaBitWidth()
            == second.GetStratumDifferentiaBitWidth()
        )
        mrca_rank_bounds = calc_rank_of_mrca_bounds_between(
            first,
            second,
            prior,
            confidence_level,
        )

        if mrca_rank_bounds is None:
            return None
        else:
            mrca_rank_lb, mrca_rank_ub = mrca_rank_bounds
            pop_mrca_rank_lower_bound = min(
                mrca_rank_lb, pop_mrca_rank_lower_bound
            )
            pop_mrca_rank_upper_bound = min(
                mrca_rank_ub, pop_mrca_rank_upper_bound
            )

    if not any_combinations:
        warnings.warn(
            "Empty or singleton population. Unable to calculate rank of MRCA "
            "bounds."
        )
        return None

    return (pop_mrca_rank_lower_bound, pop_mrca_rank_upper_bound)



import pandas as pd
import pytest

from hstrat._auxiliary_lib import alifestd_count_children_of_asexual


def sample_phylogeny_df1():
    """Fixture to provide a sample phylogeny dataframe."""
    return pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "ancestor_id": [1, 1, 1, 2],
        },
    )


def sample_phylogeny_df2():
    """Fixture to provide a sample phylogeny dataframe."""
    return pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "ancestor_list": ["[1]", "[1]", "[1]", "[2]"],
        },
    )


@pytest.mark.parametrize(
    "sample_phylogeny_df", [sample_phylogeny_df1(), sample_phylogeny_df2()]
)
def test_count_children_valid_parent(sample_phylogeny_df: pd.DataFrame):
    assert alifestd_count_children_of_asexual(sample_phylogeny_df, 1) == 2
    assert alifestd_count_children_of_asexual(sample_phylogeny_df, 2) == 1


@pytest.mark.parametrize(
    "sample_phylogeny_df", [sample_phylogeny_df1(), sample_phylogeny_df2()]
)
def test_count_children_no_children(sample_phylogeny_df: pd.DataFrame):
    assert alifestd_count_children_of_asexual(sample_phylogeny_df, 4) == 0
    assert alifestd_count_children_of_asexual(sample_phylogeny_df, 3) == 0


@pytest.mark.parametrize(
    "sample_phylogeny_df", [sample_phylogeny_df1(), sample_phylogeny_df2()]
)
def test_count_children_invalid_parent(sample_phylogeny_df: pd.DataFrame):
    with pytest.raises(ValueError):
        alifestd_count_children_of_asexual(sample_phylogeny_df, parent=999)


@pytest.mark.parametrize(
    "sample_phylogeny_df", [sample_phylogeny_df1(), sample_phylogeny_df2()]
)
def test_no_mutation_of_input(sample_phylogeny_df: pd.DataFrame):
    original_df = sample_phylogeny_df.copy()
    alifestd_count_children_of_asexual(
        sample_phylogeny_df, parent=1, mutate=False
    )
    pd.testing.assert_frame_equal(sample_phylogeny_df, original_df)
